import os
from asyncio import sleep
from contextlib import suppress
from pathlib import Path
from unittest.mock import MagicMock

import pytest
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.credentials import OAuthClientCredentials
from cognite.client.data_classes import (
    AssetWrite,
    AssetWriteList,
    ClientCredentials,
    DataPointSubscriptionWrite,
    DataSet,
    Function,
    FunctionSchedule,
    FunctionSchedulesList,
    FunctionScheduleWrite,
    FunctionScheduleWriteList,
    FunctionTaskParameters,
    FunctionWrite,
    FunctionWriteList,
    GroupWrite,
    LabelDefinitionWrite,
    TimeSeriesList,
    TimeSeriesWrite,
    TimeSeriesWriteList,
    WorkflowVersionUpsert,
    WorkflowVersionUpsertList,
    filters,
)
from cognite.client.data_classes.capabilities import IDScopeLowerCase, TimeSeriesAcl
from cognite.client.data_classes.data_modeling import NodeApplyList, NodeList, Space, ViewApply, ViewApplyList
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteFileApply, CogniteTimeSeries, CogniteTimeSeriesApply
from cognite.client.data_classes.datapoints_subscriptions import (
    DatapointSubscriptionProperty,
    DatapointSubscriptionWriteList,
)
from cognite.client.data_classes.labels import LabelDefinitionWriteList
from cognite.client.exceptions import CogniteAPIError

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.extendable_cognite_file import (
    ExtendableCogniteFileApply,
    ExtendableCogniteFileApplyList,
)
from cognite_toolkit._cdf_tk.client.data_classes.robotics import (
    DataPostProcessingWrite,
    DataPostProcessingWriteList,
    RobotCapability,
    RobotCapabilityWrite,
    RobotCapabilityWriteList,
)
from cognite_toolkit._cdf_tk.client.testing import monkeypatch_toolkit_client
from cognite_toolkit._cdf_tk.cruds import (
    AssetCRUD,
    CogniteFileCRUD,
    DataModelCRUD,
    DatapointSubscriptionCRUD,
    FunctionCRUD,
    FunctionScheduleCRUD,
    GroupCRUD,
    LabelCRUD,
    NodeCRUD,
    ResourceWorker,
    RobotCapabilityCRUD,
    RoboticsDataPostProcessingCRUD,
    TransformationCRUD,
    ViewCRUD,
    WorkflowVersionCRUD,
)
from cognite_toolkit._cdf_tk.tk_warnings import EnvironmentVariableMissingWarning, catch_warnings
from tests.test_integration.constants import RUN_UNIQUE_ID


class TestFunctionScheduleLoader:
    # The function schedule service is fairly unstable, so we need to rerun the tests if they fail.
    @pytest.mark.flaky(reruns=3, reruns_delay=10, only_rerun=["AssertionError"])
    def test_update_function_schedule(
        self,
        toolkit_client: ToolkitClient,
        toolkit_client_config: ToolkitClientConfig,
        dummy_function: Function,
        dummy_schedule: FunctionSchedule,
    ) -> None:
        loader = FunctionScheduleCRUD(toolkit_client, None, None)
        function_schedule = dummy_schedule.as_write()

        function_schedule.description = (
            "Updated description."
            if function_schedule.description != "Updated description."
            else "Original description."
        )
        identifier = loader.get_id(function_schedule)
        assert isinstance(toolkit_client_config.credentials, OAuthClientCredentials)
        loader.authentication_by_id[identifier] = ClientCredentials(
            toolkit_client_config.credentials.client_id, toolkit_client_config.credentials.client_secret
        )

        # Function schedules cannot be updated, they must be deleted and recreated.
        loader.delete([identifier])
        loader.create(FunctionScheduleWriteList([function_schedule]))

        retrieved = loader.retrieve([identifier])
        if not retrieved or retrieved[0].description != function_schedule.description:
            # The service can be a bit slow in returning the updated description,
            # so we wait a bit and try again. (Eventual consistency)
            sleep(2)
            retrieved = loader.retrieve([identifier])

        assert retrieved, "Function schedule not found after update."
        assert retrieved[0].description == function_schedule.description

    def test_creating_schedule_then_print_ids(
        self, toolkit_client: ToolkitClient, toolkit_client_config: ToolkitClientConfig, dummy_function: Function
    ) -> None:
        local = FunctionScheduleWrite(
            name="test_creating_schedule_then_print_ids",
            cron_expression="0 7 * * TUE",
            function_external_id=dummy_function.external_id,
            description="This schedule should be ignored as it does not have a function_external_id",
        )
        loader = FunctionScheduleCRUD(toolkit_client, None, None)
        assert isinstance(toolkit_client_config.credentials, OAuthClientCredentials)
        loader.authentication_by_id[loader.get_id(local)] = ClientCredentials(
            toolkit_client_config.credentials.client_id, toolkit_client_config.credentials.client_secret
        )

        created: FunctionSchedulesList | None = None
        try:
            created = loader.create(FunctionScheduleWriteList([local]))
            loader.get_ids(created)
        finally:
            if created:
                toolkit_client.functions.schedules.delete(created[0].id)


@pytest.fixture(scope="session")
def three_timeseries(cognite_client: CogniteClient) -> TimeSeriesList:
    ts_list = TimeSeriesWriteList(
        [
            TimeSeriesWrite(
                external_id=f"toolkit_test_timeseries_{i}",
                name=f"Toolkit Test TimeSeries {i}",
                is_step=False,
                is_string=False,
            )
            for i in range(1, 4)
        ]
    )
    retrieved = cognite_client.time_series.retrieve_multiple(
        external_ids=ts_list.as_external_ids(), ignore_unknown_ids=True
    )
    existing = set(retrieved.as_external_ids())
    to_create = [ts for ts in ts_list if ts.external_id not in existing]
    if to_create:
        created = cognite_client.time_series.create(to_create)
        retrieved.extend(created)
    return retrieved


@pytest.fixture(scope="session")
def one_hundred_and_one_timeseries(cognite_client: CogniteClient, toolkit_dataset: DataSet) -> TimeSeriesList:
    ts_list = TimeSeriesWriteList(
        [
            TimeSeriesWrite(
                external_id=f"toolkit_101_timeseries_{i}",
                name=f"Toolkit Test 101 TimeSeries {i}",
                is_step=False,
                is_string=False,
                data_set_id=toolkit_dataset.id,
            )
            for i in range(1, 102)
        ]
    )
    retrieved = cognite_client.time_series.retrieve_multiple(
        external_ids=ts_list.as_external_ids(), ignore_unknown_ids=True
    )
    existing = set(retrieved.as_external_ids())
    to_create = [ts for ts in ts_list if ts.external_id not in existing]
    if to_create:
        created = cognite_client.time_series.create(to_create)
        retrieved.extend(created)
    return retrieved


@pytest.fixture(scope="session")
def three_hundred_and_three_cognite_timeseries(
    toolkit_client: ToolkitClient, toolkit_space: Space
) -> NodeList[CogniteTimeSeries]:
    ts_list = NodeApplyList(
        [
            CogniteTimeSeriesApply(
                space=toolkit_space.space,
                external_id=f"toolkit_303_test_timeseries_{i}",
                is_step=False,
                time_series_type="numeric",
            )
            for i in range(1, 304)
        ]
    )
    retrieved = toolkit_client.data_modeling.instances.retrieve_nodes(ts_list.as_ids(), node_cls=CogniteTimeSeries)
    if len(retrieved) == len(ts_list):
        # All timeseries already exist, return the retrieved ones
        return retrieved
    existing = set(retrieved.as_ids())
    to_create = [ts for ts in ts_list if ts.as_id() not in existing]
    if to_create:
        _ = toolkit_client.data_modeling.instances.apply(to_create)
    return toolkit_client.data_modeling.instances.retrieve_nodes(ts_list.as_ids(), node_cls=CogniteTimeSeries)


class TestDatapointSubscriptionLoader:
    def test_delete_non_existing(self, toolkit_client: ToolkitClient) -> None:
        loader = DatapointSubscriptionCRUD(toolkit_client, None)
        delete_count = loader.delete(["non_existing"])
        assert delete_count == 0

    def test_create_update_delete_subscription(self, toolkit_client: ToolkitClient) -> None:
        sub = DataPointSubscriptionWrite(
            external_id=f"tmp_test_create_update_delete_subscription_{RUN_UNIQUE_ID}",
            partition_count=1,
            name="Initial name",
            filter=filters.Prefix(DatapointSubscriptionProperty.external_id, "ts_value"),
        )
        update = DataPointSubscriptionWrite(
            external_id=f"tmp_test_create_update_delete_subscription_{RUN_UNIQUE_ID}",
            partition_count=1,
            name="Updated name",
            filter=filters.Prefix(DatapointSubscriptionProperty.external_id, "ts_value"),
        )

        loader = DatapointSubscriptionCRUD(toolkit_client, None)

        try:
            created = loader.create(DatapointSubscriptionWriteList([sub]))
            assert len(created) == 1

            updated = loader.update(DatapointSubscriptionWriteList([update]))
            assert len(updated) == 1
            assert updated[0].name == "Updated name"
        finally:
            loader.delete([sub.external_id])

    def test_create_update_delete_subscription_with_ids(
        self,
        toolkit_client: ToolkitClient,
        one_hundred_and_one_timeseries: TimeSeriesList,
        three_timeseries: TimeSeriesList,
        three_hundred_and_three_cognite_timeseries: NodeList[CogniteTimeSeries],
    ) -> None:
        ts_ids = "\n- ".join(one_hundred_and_one_timeseries.as_external_ids())
        cognite_ts_ids = "\n".join(
            [
                f"- space: {node.space}\n  externalId: {node.external_id}"
                for node in three_hundred_and_three_cognite_timeseries
            ]
        )
        sub_yaml = f"""externalId: tmp_test_create_update_delete_101_subscription_{RUN_UNIQUE_ID}
partitionCount: 1
name: The subscription name
timeSeriesIds:
- {ts_ids}
instanceIds:
{cognite_ts_ids}
"""
        ts_update_ds = "\n- ".join(
            one_hundred_and_one_timeseries.as_external_ids() + three_timeseries.as_external_ids()
        )
        update_yaml = f"""externalId: tmp_test_create_update_delete_101_subscription_{RUN_UNIQUE_ID}
partitionCount: 1
name: The subscription name
timeSeriesIds:
- {ts_update_ds}
"""
        loader = DatapointSubscriptionCRUD(toolkit_client, None)
        sub = self._load_subscription_from_yaml(self._create_mock_file(sub_yaml), loader)
        try:
            created = loader.create(DatapointSubscriptionWriteList([sub]))
            assert len(created) == 1
            initial_description = created[0].description
            assert created[0].time_series_count == len(one_hundred_and_one_timeseries) + len(
                three_hundred_and_three_cognite_timeseries
            ), "The subscription should have the correct number of time series"

            update = self._load_subscription_from_yaml(self._create_mock_file(update_yaml), loader)
            updated = loader.update(DatapointSubscriptionWriteList([update]))
            assert len(updated) == 1
            updated_description = updated[0].description
            assert updated_description != initial_description, (
                "The description should have changed after the update with a hash fo the timeseries IDs"
            )
            assert updated[0].time_series_count == len(one_hundred_and_one_timeseries) + len(three_timeseries), (
                "The subscription should have the correct number of time series after the update"
            )
        finally:
            loader.delete([sub.external_id])

    def test_no_redeploy_ids_defined(
        self, toolkit_client: ToolkitClient, toolkit_dataset: DataSet, three_timeseries: TimeSeriesList
    ) -> None:
        definition_yaml = f"""externalId: toolkit_test_no_redeploy_ids_defined
name: Test no redeploy IDs defined
partitionCount: 1
dataSetExternalId: {toolkit_dataset.external_id}
timeSeriesIds:
- {three_timeseries[0].external_id}
- {three_timeseries[1].external_id}
- {three_timeseries[2].external_id}
"""
        loader = DatapointSubscriptionCRUD.create_loader(toolkit_client)

        filepath = self._create_mock_file(definition_yaml)
        resource = self._load_subscription_from_yaml(filepath, loader)
        assert isinstance(resource, DataPointSubscriptionWrite)
        if not loader.retrieve([resource.external_id]):
            _ = loader.create(DatapointSubscriptionWriteList([resource]))

        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([filepath])

        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}

    @staticmethod
    def _create_mock_file(yaml_content: str) -> Path:
        mock_file = MagicMock(spec=Path)
        mock_file.read_text.return_value = yaml_content
        return mock_file

    @staticmethod
    def _load_subscription_from_yaml(filepath: Path, loader: DatapointSubscriptionCRUD) -> DataPointSubscriptionWrite:
        resource_dict = loader.load_resource_file(filepath, {})
        assert len(resource_dict) == 1
        return loader.load_resource(resource_dict[0])


class TestLabelLoader:
    def test_delete_non_existing(self, cognite_client: CogniteClient) -> None:
        loader = LabelCRUD(cognite_client, None)
        delete_count = loader.delete(["non_existing"])
        assert delete_count == 0

    def test_create_delete_label(self, toolkit_client: ToolkitClient) -> None:
        label = LabelDefinitionWrite(
            external_id=f"tmp_test_create_update_delete_label_{RUN_UNIQUE_ID}",
            name="Initial name",
        )
        loader = LabelCRUD(toolkit_client, None)

        try:
            created = loader.create(LabelDefinitionWriteList([label]))
            assert len(created) == 1
        finally:
            loader.delete([label.external_id])


class TestAssetLoader:
    def test_create_delete_asset(self, cognite_client: CogniteClient) -> None:
        asset = AssetWrite(
            external_id=f"tmp_test_create_delete_asset_{RUN_UNIQUE_ID}",
            name="My Asset",
            description="My description",
        )

        loader = AssetCRUD(cognite_client, None)

        try:
            created = loader.create(AssetWriteList([asset]))
            assert len(created) == 1

            delete_count = loader.delete([asset.external_id])
            assert delete_count == 1
        finally:
            # Ensure that the asset is deleted even if the test fails.
            cognite_client.assets.delete(external_id=asset.external_id, ignore_unknown_ids=True)


@pytest.fixture
def existing_robot_capability(toolkit_client: ToolkitClient) -> RobotCapability:
    write = RobotCapabilityWrite(
        name="integration_test_robot_capability",
        description="Test robot capability",
        external_id="integration_test_robot_capability",
        method="ptz",
        input_schema={},
        data_handling_schema={},
    )

    try:
        return toolkit_client.robotics.capabilities.retrieve(write.external_id)
    except CogniteAPIError:
        return toolkit_client.robotics.capabilities.create(write)


class TestRobotCapability:
    def test_retrieve_existing_and_not_existing(
        self, toolkit_client: ToolkitClient, existing_robot_capability: RobotCapability
    ) -> None:
        loader = RobotCapabilityCRUD(toolkit_client, None)

        capabilities = loader.retrieve([existing_robot_capability.external_id, "non_existing_robot"])

        assert len(capabilities) == 1

    def test_create_update_retrieve_delete(self, toolkit_client: ToolkitClient) -> None:
        loader = RobotCapabilityCRUD(toolkit_client, None)

        original = RobotCapabilityWrite.load("""name: Read dial gauge
externalId: read_dial_gauge
method: read_dial_gauge
description: Original Description
inputSchema:
    $schema: http://json-schema.org/draft-07/schema#
    id: robotics/schemas/0.1.0/capabilities/ptz
    title: PTZ camera capability input
dataHandlingSchema:
    $schema: http://json-schema.org/draft-07/schema#
    id: robotics/schemas/0.1.0/data_handling/read_dial_gauge
    title: Read dial gauge data handling
""")

        update = RobotCapabilityWrite.load("""name: Read dial gauge
externalId: read_dial_gauge
method: read_dial_gauge
description: Original Description
inputSchema:
    $schema: http://json-schema.org/draft-07/schema#
    id: robotics/schemas/0.2.0/capabilities/ptz
    title: Updated PTZ camera capability input
dataHandlingSchema:
    $schema: http://json-schema.org/draft-07/schema#
    id: robotics/schemas/0.2.0/data_handling/read_dial_gauge
    title: Updated read dial gauge data handling
""")
        try:
            created = loader.create(RobotCapabilityWriteList([original]))
            assert len(created) == 1

            updated = loader.update(RobotCapabilityWriteList([update]))
            assert len(updated) == 1
            assert updated[0].input_schema == update.input_schema

            retrieved = loader.retrieve([original.external_id])
            assert len(retrieved) == 1
            assert retrieved[0].input_schema == update.input_schema
        finally:
            loader.delete([original.external_id])


class TestRobotDataPostProcessing:
    def test_create_update_retrieve_delete(self, toolkit_client: ToolkitClient) -> None:
        loader = RoboticsDataPostProcessingCRUD(toolkit_client, None)

        original = DataPostProcessingWrite.load("""name: Read dial gauge
externalId: read_dial_gauge
method: read_dial_gauge
description: Original Description
inputSchema:
  $schema: http://json-schema.org/draft-07/schema#
  id: robotics/schemas/0.1.0/capabilities/ptz
  title: PTZ camera capability input
  type: object
  properties:
    method:
      type: string
    parameters:
      type: object
      properties:
        tilt:
          type: number
          minimum: -90
          maximum: 90
        pan:
          type: number
          minimum: -180
          maximum: 180
        zoom:
          type: number
          minimum: 0
          maximum: 100
      required:
      - tilt
      - pan
      - zoom
  required:
  - method
  - parameters
  additionalProperties: false
""")

        update = DataPostProcessingWrite.load("""method: read_dial_gauge
name: Read dial gauge
externalId: read_dial_gauge
description: Read dial gauge from an image using Cognite Vision gauge reader
inputSchema:
  $schema: http://json-schema.org/draft-07/schema#
  id: robotics/schemas/0.1.0/data_postprocessing/read_dial_gauge
  title: Read dial gauge input
  type: object
  properties:
    image:
      type: object
      properties:
        method:
          type: string
        parameters:
          type: object
          properties:
            unit:
              type: string
            deadAngle:
              type: number
            minLevel:
              type: number
            maxLevel:
              type: number
      required:
        - method
        - parameters
      additionalProperties: false
  additionalProperties: false""")

        # Ensure the original is deleted even if the test fails
        loader.delete([original.external_id])

        try:
            created = loader.create(DataPostProcessingWriteList([original]))
            assert len(created) == 1

            updated = loader.update(DataPostProcessingWriteList([update]))
            assert len(updated) == 1
            assert updated[0].input_schema == update.input_schema

            retrieved = loader.retrieve([original.external_id])
            assert len(retrieved) == 1
            assert retrieved[0].input_schema == update.input_schema
        finally:
            loader.delete([original.external_id])


@pytest.fixture(scope="module")
def a_container(toolkit_client: ToolkitClient, toolkit_space: dm.Space) -> dm.Container:
    return toolkit_client.data_modeling.containers.apply(
        dm.ContainerApply(
            name=f"container_test_resource_loaders_{RUN_UNIQUE_ID}",
            space=toolkit_space.space,
            external_id=f"container_test_resource_loaders_{RUN_UNIQUE_ID}",
            properties={"name": dm.ContainerProperty(type=dm.Text())},
        )
    )


@pytest.fixture(scope="module")
def two_views(toolkit_client: ToolkitClient, toolkit_space: dm.Space, a_container: dm.Container) -> dm.ViewList:
    return toolkit_client.data_modeling.views.apply(
        [
            dm.ViewApply(
                space=toolkit_space.space,
                external_id=f"first_view{RUN_UNIQUE_ID}",
                version="1",
                properties={
                    "name": dm.MappedPropertyApply(container=a_container.as_id(), container_property_identifier="name")
                },
            ),
            dm.ViewApply(
                space=toolkit_space.space,
                external_id=f"second_view{RUN_UNIQUE_ID}",
                version="1",
                properties={
                    "alsoName": dm.MappedPropertyApply(
                        container=a_container.as_id(), container_property_identifier="name", name="name2"
                    )
                },
            ),
        ]
    )


class TestDataModelLoader:
    def test_create_update_delete(
        self, toolkit_client: ToolkitClient, toolkit_space: dm.Space, two_views: dm.ViewList
    ) -> None:
        loader = DataModelCRUD(toolkit_client, None)
        view_list = two_views.as_ids()
        assert len(view_list) == 2, "Expected 2 views in the test data model"
        my_model = dm.DataModelApply(
            name="My model",
            description="Original description",
            views=view_list,
            space=toolkit_space.space,
            external_id=f"tmp_test_create_update_delete_data_model_{RUN_UNIQUE_ID}",
            version="1",
        )

        try:
            created = loader.create(dm.DataModelApplyList([my_model]))
            assert len(created) == 1

            update = dm.DataModelApply.load(my_model.dump())
            update.views = [view_list[0]]

            with pytest.raises(CogniteAPIError):
                loader.update(dm.DataModelApplyList([update]))
            # You need to update the version to update the model
            update.version = "2"

            updated = loader.update(dm.DataModelApplyList([update]))
            assert len(updated) == 1
            assert updated[0].views == [view_list[0]]
        finally:
            loader.delete([my_model.as_id()])


@pytest.fixture
def custom_file_container(toolkit_client: ToolkitClient, toolkit_space: dm.Space) -> dm.Container:
    return toolkit_client.data_modeling.containers.apply(
        dm.ContainerApply(
            name=f"container_test_resource_loaders_{RUN_UNIQUE_ID}",
            space=toolkit_space.space,
            external_id=f"container_test_resource_loaders_{RUN_UNIQUE_ID}",
            properties={
                "status": dm.ContainerProperty(type=dm.Text()),
                "fileCategory": dm.ContainerProperty(type=dm.Text()),
            },
        )
    )


@pytest.fixture
def cognite_file_extension(
    toolkit_client: ToolkitClient, custom_file_container: dm.Container, toolkit_space: dm.Space
) -> dm.View:
    container = custom_file_container.as_id()
    return toolkit_client.data_modeling.views.apply(
        dm.ViewApply(
            space=toolkit_space.space,
            external_id="CogniteFileExtension",
            version="v1",
            implements=[CogniteFileApply.get_source()],
            properties={
                "status": dm.MappedPropertyApply(container=container, container_property_identifier="status"),
                "fileCategory": dm.MappedPropertyApply(
                    container=container, container_property_identifier="fileCategory"
                ),
            },
        )
    )


class TestCogniteFileLoader:
    def test_create_update_retrieve_delete(self, toolkit_client: ToolkitClient, toolkit_space: dm.Space) -> None:
        loader = CogniteFileCRUD(toolkit_client, None)
        # Loading from YAML to test the loading of extra properties as well
        file = ExtendableCogniteFileApply.load(f"""space: {toolkit_space.space}
externalId: tmp_test_create_update_delete_file_{RUN_UNIQUE_ID}
name: My file
description: Original description
""")
        try:
            created = loader.create(ExtendableCogniteFileApplyList([file]))
            assert len(created) == 1

            retrieved = loader.retrieve([file.as_id()])
            assert len(retrieved) == 1
            assert retrieved[0].name == "My file"
            assert retrieved[0].description == "Original description"

            update = ExtendableCogniteFileApply._load(file.dump(context="local"))
            update.description = "Updated description"

            updated = loader.update(ExtendableCogniteFileApplyList([update]))
            assert len(updated) == 1

            retrieved = loader.retrieve([file.as_id()])
            assert len(retrieved) == 1
            assert retrieved[0].description == "Updated description"
            assert retrieved[0].name == "My file"
        finally:
            loader.delete([file.as_id()])

    @pytest.mark.skip("For now, we do not support creating extensions")
    def test_create_update_retrieve_delete_extension(
        self, toolkit_client: ToolkitClient, cognite_file_extension: dm.View, toolkit_space: dm.Space
    ) -> None:
        loader = CogniteFileCRUD(toolkit_client, None)
        view_id = cognite_file_extension.as_id()
        # Loading from YAML to test the loading of extra properties as well
        file = ExtendableCogniteFileApply.load(f"""space: {toolkit_space.space}
externalId: tmp_test_create_update_delete_file_extension_{RUN_UNIQUE_ID}
name: MyExtendedFile
description: Original description
nodeSource:
  space: {view_id.space}
  externalId: {view_id.external_id}
  version: {view_id.version}
  type: view
status: Active
fileCategory: Document
""")
        try:
            created = loader.create(ExtendableCogniteFileApplyList([file]))
            assert len(created) == 1

            update = ExtendableCogniteFileApply._load(file.dump(context="local"))
            # Ensure serialization and deserialization works
            assert update.name == "MyExtendedFile"
            assert update.extra_properties is not None
            assert update.extra_properties["fileCategory"] == "Document"
            update.extra_properties["status"] = "Inactive"

            updated = loader.update(ExtendableCogniteFileApplyList([update]))
            assert len(updated) == 1

            retrieved = loader.retrieve([file.as_id()])
            assert len(retrieved) == 1
            assert retrieved[0].name == "MyExtendedFile"
            assert retrieved[0].extra_properties is not None
            assert retrieved[0].extra_properties["status"] == "Inactive"
        finally:
            loader.delete([file.as_id()])


class TestGroupLoader:
    def test_dump_cdf_group_with_invalid_reference(self, toolkit_client: ToolkitClient) -> None:
        to_delete = TimeSeriesWrite(
            external_id="test_dump_cdf_group_with_invalid_reference",
            name="Test dump CDF group with invalid reference",
            is_step=False,
            is_string=False,
        )
        group_id: int | None = None
        try:
            created_ts = toolkit_client.time_series.create(to_delete)
            group = GroupWrite(
                name="test_dump_cdf_group_with_invalid_reference",
                source_id="1234-dummy",
                capabilities=[
                    TimeSeriesAcl(actions=[TimeSeriesAcl.Action.Read], scope=IDScopeLowerCase([created_ts.id]))
                ],
            )
            created_group = toolkit_client.iam.groups.create(group)
            group_id = created_group.id
            toolkit_client.time_series.delete(id=created_ts.id)

            loader = GroupCRUD.create_loader(toolkit_client)

            dumped = loader.dump_resource(created_group)
            assert "capabilities" in dumped
            capabilities = dumped["capabilities"]
            assert isinstance(capabilities, list)
            assert len(capabilities) == 1
            assert capabilities[0] == {"timeSeriesAcl": {"actions": ["READ"], "scope": {"idscope": {"ids": []}}}}
        finally:
            toolkit_client.time_series.delete(external_id=to_delete.external_id, ignore_unknown_ids=True)
            if group_id:
                with suppress(CogniteAPIError):
                    toolkit_client.iam.groups.delete(id=group_id)


class TestWorkflowVersionLoader:
    def test_load_task_with_reference(self) -> None:
        definition_yaml = """workflowExternalId: myWorkflow
version: v1
workflowDefinition:
  description: Two tasks with reference
  tasks:
  - externalId: myTask2
    type: function
    parameters:
      function:
        externalId: fn_first_function
  - externalId: myTask2
    type: function
    parameters:
      function:
        externalId: ${myTask1.output.nextFunction}
        data: ${myTask1.output.data}
    dependsOn:
    - externalId: myTask1
"""
        file = MagicMock(spec=Path)
        file.read_text.return_value = definition_yaml
        with monkeypatch_toolkit_client() as client:
            loader = WorkflowVersionCRUD(client, None, None)

            with catch_warnings(EnvironmentVariableMissingWarning) as warning_list:
                loaded = loader.load_resource_file(file, {"myTask1.output.data": "should-be-ignored"})
            definition = loader.load_resource(loaded[0])

        assert len(definition.workflow_definition.tasks) == 2
        task2 = definition.workflow_definition.tasks[1]
        parameters = task2.parameters
        assert isinstance(parameters, FunctionTaskParameters)
        assert parameters.data == "${myTask1.output.data}"
        assert len(warning_list) == 0, "We should not get a warning for using a reference in a task parameter"

    def test_load_workflow_without_defaults_not_redeployed(self, toolkit_client: ToolkitClient) -> None:
        definition_yaml = """workflowExternalId: testWorkflowWithoutDefaults
version: v1
workflowDefinition:
  description: Tasks without defaults
  tasks:
  - externalId: myTask1
    type: function
    parameters:
      function:
        externalId: fn_first_function
  - externalId: myTask2
    type: transformation
    parameters:
      transformation:
        externalId: some_transformation
    retries: null
"""
        loader = WorkflowVersionCRUD.create_loader(toolkit_client)

        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = definition_yaml

        resource_dict = loader.load_resource_file(filepath, {})
        assert len(resource_dict) == 1
        resource = loader.load_resource(resource_dict[0])
        assert isinstance(resource, WorkflowVersionUpsert)
        if not loader.retrieve([resource.as_id()]):
            _ = loader.create(WorkflowVersionUpsertList([resource]))

        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([filepath])

        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}


class TestTransformationLoader:
    def test_create_transformation_auth_without_scope(self, toolkit_client: ToolkitClient) -> None:
        transformation_text = """externalId: transformation_without_scope
name: This is a test transformation
destination:
  type: assets
ignoreNullFields: true
isPublic: true
conflictMode: upsert
query: Select * from assets
# Reusing the credentials from the Toolkit principal
authentication:
  clientId: ${IDP_CLIENT_ID}
  clientSecret: ${IDP_CLIENT_SECRET}
"""
        loader = TransformationCRUD.create_loader(toolkit_client)
        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = transformation_text

        loaded = loader.load_resource_file(filepath, dict(os.environ))
        assert len(loaded) == 1
        transformation = loader.load_resource(loaded[0])

        try:
            created = loader.create([transformation])
            assert len(created) == 1
        finally:
            toolkit_client.transformations.delete(external_id="transformation_without_scope", ignore_unknown_ids=True)

    def test_create_transformation_reusing_source_destination_auth(self, toolkit_client: ToolkitClient) -> None:
        transformation_text = """externalId: transformation_reusing_source_destination_auth
name: This is a test transformation from the Toolkit
destination:
  type: assets
ignoreNullFields: true
isPublic: true
conflictMode: upsert
query: Select * from assets
# Reusing the credentials from the Toolkit principal
authentication:
  read:
    clientId: ${IDP_CLIENT_ID}
    clientSecret: ${IDP_CLIENT_SECRET}
  write:
    clientId: ${IDP_CLIENT_ID}
    clientSecret: ${IDP_CLIENT_SECRET}
        """
        loader = TransformationCRUD.create_loader(toolkit_client)
        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = transformation_text

        loaded = loader.load_resource_file(filepath, dict(os.environ))
        assert len(loaded) == 1
        transformation = loader.load_resource(loaded[0])

        try:
            created_list = loader.create([transformation])
            assert len(created_list) == 1
            created = created_list[0]
            assert created.source_session is not None
            assert created.destination_session is not None
            assert created.source_session.session_id != created.destination_session.session_id, (
                "There should be different sessions for source and destination authentication even"
                " if they reuse the same credentials"
            )
        finally:
            toolkit_client.transformations.delete(
                external_id="transformation_reusing_source_destination_auth", ignore_unknown_ids=True
            )

    @pytest.mark.skip(
        reason="This is a load test that takes a long time to run ~5 minutes, "
        "and puts a high load on the transformation service. "
        "It is used to verify that a fix works, but not run regularly."
    )
    def test_load_test_transformation_creation(self, toolkit_client: ToolkitClient) -> None:
        credentials = toolkit_client.config.credentials
        if not isinstance(credentials, OAuthClientCredentials):
            pytest.skip("This test requires OAuthClientCredentials to run")
        secret = credentials.client_secret
        client_id = credentials.client_id
        N = 500
        definition_yaml = [
            f"""- externalId: test_load_transformation_creation_{RUN_UNIQUE_ID}_{i}
  name: Load Test Transformation Creation {i}
  destination:
    type: assets
  ignoreNullFields: true
  isPublic: true
  conflictMode: upsert
  query: Select * from assets
  authentication:
    clientId: {client_id}
    clientSecret: {secret}
"""
            for i in range(1, N + 1)
        ]

        loader = TransformationCRUD.create_loader(toolkit_client)
        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = "\n".join(definition_yaml)

        loaded = loader.load_resource_file(filepath, {})
        assert len(loaded) == N
        transformations = [loader.load_resource(resource) for resource in loaded]

        try:
            created_list = loader.create(transformations)
            assert len(created_list) == N
        finally:
            loader.delete([transformation.external_id for transformation in transformations])


class TestNodeLoader:
    def test_update_existing_node(self, toolkit_client: ToolkitClient, toolkit_space: dm.Space) -> None:
        loader = NodeCRUD(toolkit_client, None)
        view_id = dm.ViewId("cdf_cdm", "CogniteDescribable", "v1")
        existing_node = dm.NodeApply(
            space=toolkit_space.space,
            external_id=f"toolkit_test_update_existing_node_{RUN_UNIQUE_ID}",
            sources=[
                dm.NodeOrEdgeData(
                    view_id,
                    {
                        "name": "existing name",
                        "description": "Existing description",
                    },
                )
            ],
        )
        updated_node = dm.NodeApply(
            space=existing_node.space,
            external_id=existing_node.external_id,
            sources=[
                dm.NodeOrEdgeData(
                    view_id,
                    {
                        "name": "updated name",
                        "aliases": ["alias1", "alias2"],
                    },
                )
            ],
        )
        try:
            created = loader.create(dm.NodeApplyList([existing_node]))
            assert len(created) == 1

            updated = loader.update(dm.NodeApplyList([updated_node]))
            assert len(updated) == 1

            retrieved = toolkit_client.data_modeling.instances.retrieve(existing_node.as_id(), sources=[view_id])
            assert len(retrieved.nodes) == 1
            node = retrieved.nodes[0]
            assert node.properties[view_id] == {
                "name": "updated name",  # Overwrite
                "description": "Existing description",  # Keep existing description
                "aliases": ["alias1", "alias2"],  # Add new aliases
            }
        finally:
            loader.delete([existing_node.as_id()])


class TestViewLoader:
    def test_no_implement_not_redeployed(
        self, toolkit_client: ToolkitClient, toolkit_space: dm.Space, a_container: dm.Container
    ) -> None:
        definition_yaml = f"""space: {toolkit_space.space}
externalId: ToolkitTestNoImplementsNotRedeployed{RUN_UNIQUE_ID}
version: v1
implements: []
properties:
  name:
    container:
      space: {a_container.space}
      externalId: {a_container.external_id}
      type: container
    containerPropertyIdentifier: name
        """
        loader = ViewCRUD.create_loader(toolkit_client)

        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = definition_yaml

        resource_dict = loader.load_resource_file(filepath, {})
        assert len(resource_dict) == 1
        resource = loader.load_resource(resource_dict[0])
        assert isinstance(resource, ViewApply)
        if not loader.retrieve([resource.as_id()]):
            _ = loader.create(ViewApplyList([resource]))

        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([filepath])

        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}


class TestFunctionLoader:
    def test_avoid_redeploying_function_with_no_changes(
        self, toolkit_client: ToolkitClient, toolkit_dataset: DataSet, tmp_path: Path
    ) -> None:
        function_code = """from cognite.client import CogniteClient


def handle(data: dict, client: CogniteClient, secrets: dict, function_call_info: dict) -> dict:
    # This will fail unless the function has the specified capabilities.
    print("Print statements will be shown in the logs.")
    print("Running with the following configuration:\n")
    return {
        "data": data,
        "functionInfo": function_call_info,
    }

"""
        external_id = "toolkit_test_function_no_redeploy"
        definition_yaml = f"""externalId: {external_id}
name: Toolkit Test Function No Redeploy
owner: ""
dataSetExternalId: {toolkit_dataset.external_id}
description: ""
        """
        build_dir = tmp_path / "build"
        function_code_path = build_dir / FunctionCRUD.folder_name / external_id / "handler.py"
        function_code_path.parent.mkdir(parents=True, exist_ok=True)
        function_code_path.write_text(function_code, encoding="utf-8")

        loader = FunctionCRUD.create_loader(toolkit_client, build_dir)
        filepath = MagicMock(spec=Path)
        filepath.read_text.return_value = definition_yaml
        filepath.parent.name = FunctionCRUD.folder_name
        resource_dict = loader.load_resource_file(filepath, {})
        assert len(resource_dict) == 1
        resource = loader.load_resource(resource_dict[0])
        assert isinstance(resource, FunctionWrite)
        if not loader.retrieve([resource.external_id]):
            _ = loader.create(FunctionWriteList([resource]))
        worker = ResourceWorker(loader, "deploy")
        resources = worker.prepare_resources([filepath])
        assert {
            "create": len(resources.to_create),
            "change": len(resources.to_update),
            "delete": len(resources.to_delete),
            "unchanged": len(resources.unchanged),
        } == {"create": 0, "change": 0, "delete": 0, "unchanged": 1}



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.cognitefile import CogniteFileYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def valid_cognitefile_test_cases() -> Iterable:
    yield pytest.param({"space": "my_space", "externalId": "my_file_123"}, id="minimal-valid-file")
    yield pytest.param(
        {
            "space": "production_space",
            "externalId": "complete_file_456",
            "name": "Complete Test File",
            "description": "Test file with all properties",
            "tags": ["test", "documentation", "example"],
            "aliases": ["test_file", "sample_file"],
            "sourceId": "source_123",
            "sourceContext": "production_system",
            "source": {"space": "source_space", "externalId": "source_ref_123"},
            "sourceCreatedTime": "2025-01-15T10:30:00Z",
            "sourceUpdatedTime": "2025-08-01T15:45:30Z",
            "sourceCreatedUser": "user123",
            "sourceUpdatedUser": "user456",
            "assets": [
                {"space": "asset_space", "externalId": "asset_1"},
                {"space": "asset_space", "externalId": "asset_2"},
            ],
            "mimeType": "application/pdf",
            "directory": "/documents/reports/",
            "category": {"space": "category_space", "externalId": "document_category"},
            "existingVersion": 3,
            "type": {"space": "type_space", "externalId": "report_type"},
            "nodeSource": {"type": "view", "space": "view_space", "externalId": "file_view", "version": "v1"},
            "extraProperties": {"custom_field": "custom_value", "priority": 5, "is_public": True},
        },
        id="complete-file-with-all-fields",
    )
    yield pytest.param(
        {
            "space": "test_space",
            "externalId": "data_types_test",
            "name": "Data Types Test",
            "existingVersion": 0,
            "extraProperties": {"number": 42, "boolean": False, "nested": {"key": "value"}},
        },
        id="file-with-various-data-types",
    )


def invalid_cognitefile_test_cases() -> Iterable:
    yield pytest.param({"externalId": "my_file"}, {"Missing required field: 'space'"}, id="missing-required-field")
    yield pytest.param(
        {"space": "my_space", "externalId": "my_file", "sourceCreatedTime": "invalid_date"},
        {"In field sourceCreatedTime input should be a valid datetime. Got 'invalid_date' of type str."},
        id="invalid-sourceCreatedTime-format",
    )
    yield pytest.param(
        {"space": "my_space", "externalId": "my_file", "source": {"space": "source_space"}},
        {"In source missing required field: 'externalId'"},
        id="invalid-DirectRelationReference-missing-externalId",
    )
    yield pytest.param(
        {
            "space": "my_space",
            "externalId": "my_file",
            "nodeSource": {"type": "view", "space": "view_space", "externalId": "file_view"},
        },
        {"In nodeSource missing required field: 'version'"},
        id="invalid-ViewReference-missing-version",
    )
    yield pytest.param(
        {"space": "my_space", "externalId": "my_file", "existingVersion": "not_a_number"},
        {"In field existingVersion input should be a valid integer. Got 'not_a_number' of type str."},
        id="invalid-existingVersion-type",
    )
    yield pytest.param(
        {"space": "my_space", "externalId": "a" * 256},
        {"In field externalId string should have at most 255 characters"},
        id="invalid-externalId-too-long",
    )


class TestCogniteFileYAML:
    @pytest.mark.parametrize("data", list(find_resources("CogniteFile")))
    def test_load_valid_cognitefile_file(self, data: dict) -> None:
        loaded = CogniteFileYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize(
        "data",
        [
            *valid_cognitefile_test_cases(),
            pytest.param(
                {
                    "space": "my_space",
                    "externalId": "$FILENAME",
                    "mimeType": "application/text",
                },
                id="CogniteFile with placeholder",
            ),
        ],
    )
    def test_load_valid_cognitefile(self, data: dict) -> None:
        loaded = CogniteFileYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True, mode="json") == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_cognitefile_test_cases()))
    def test_invalid_cognitefile(self, data: dict, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, CogniteFileYAML, Path("test_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert set(format_warning.errors) == expected_errors



from typing import Literal

from pydantic import Field

from cognite_toolkit._cdf_tk.constants import SPACE_FORMAT_PATTERN

from .base import ToolkitResource


class FunctionsYAML(ToolkitResource):
    external_id: str = Field(
        description="The external ID provided by the client.",
        max_length=255,
    )
    name: str = Field(
        description="The name of the function.",
        max_length=140,
    )
    description: str | None = Field(
        default=None,
        description="The description of the function.",
        min_length=1,
        max_length=500,
    )
    owner: str | None = Field(
        default=None,
        description="Owner of the function.",
        max_length=128,
    )
    function_path: str | None = Field(
        default=None,
        description="Relative path from the root folder to the file containing the handle function.",
        max_length=500,
    )
    secrets: dict[str, str] | None = Field(
        default=None,
        description="Secrets attached to the function.",
        max_length=30,
    )
    env_vars: dict[str, str] | None = Field(
        default=None,
        description="User specified environment variables on the function.",
        max_length=100,
    )
    cpu: float | None = Field(default=None, description="Number of CPU cores per function.")
    memory: float | None = Field(default=None, description="Memory per function measured in GB.")
    runtime: Literal["py39", "py310", "py311"] | None = Field(default="py311", description="Runtime of the function.")
    metadata: dict[str, str] | None = Field(
        default=None, description="Custom, application-specific metadata.", max_length=16
    )
    index_url: str | None = Field(
        default=None,
        description="Specify a different python package index, allowing for packages published in private repositories.",
    )
    extra_index_urls: list[str] | None = Field(
        default=None,
        description=" Extra package index URLs to use when building the function, allowing for packages published in private repositories.",
    )
    data_set_external_id: str | None = Field(
        default=None,
        description="Target dataset external ID for the file where the function code is stored.",
        max_length=255,
    )
    space: str | None = Field(
        default=None,
        description="The space for the CogniteFile where the function code is stored.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )



import sys
from abc import ABC
from collections import UserDict
from collections.abc import Iterable
from dataclasses import dataclass
from functools import total_ordering
from typing import Literal

from rich.table import Table

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self


@total_ordering
@dataclass
class DeployResult(ABC):
    name: str

    def __lt__(self, other: object) -> bool:
        if isinstance(other, DeployResult):
            return self.name < other.name
        else:
            return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DeployResult):
            return self.name == other.name
        else:
            return NotImplemented


@dataclass
class ResourceDeployResult(DeployResult):
    created: int = 0
    deleted: int = 0
    changed: int = 0
    unchanged: int = 0
    total: int = 0

    @property
    def calculated_total(self) -> int:
        return self.created + self.deleted + self.changed + self.unchanged

    def __iadd__(self, other: "ResourceDeployResult") -> "ResourceDeployResult":
        if self.name != other.name:
            raise ValueError("Cannot add two DeployResult objects with different names")
        self.created += other.created
        self.deleted += other.deleted
        self.changed += other.changed
        self.unchanged += other.unchanged
        self.total += other.total

        if isinstance(other, ResourceContainerDeployResult):
            return ResourceContainerDeployResult(
                name=self.name,
                created=self.created,
                deleted=self.deleted,
                changed=self.changed,
                unchanged=self.unchanged,
                total=self.total,
                item_name=other.item_name,
                dropped_datapoints=other.dropped_datapoints,
            )
        else:
            return self


@dataclass
class ResourceContainerDeployResult(ResourceDeployResult):
    item_name: str = ""
    dropped_datapoints: int = 0

    def __iadd__(self, other: ResourceDeployResult) -> "ResourceContainerDeployResult":
        if self.name != other.name:
            raise ValueError("Cannot add two ResourceContainerDeployResult objects with different names")
        super().__iadd__(other)
        if isinstance(other, ResourceContainerDeployResult):
            self.dropped_datapoints += other.dropped_datapoints
        return self

    @classmethod
    def from_resource_deploy_result(
        cls, result: ResourceDeployResult, item_name: str = "", dropped_datapoints: int = 0
    ) -> Self:
        return cls(
            name=result.name,
            created=result.created,
            deleted=result.deleted,
            changed=result.changed,
            unchanged=result.unchanged,
            total=result.total,
            item_name=item_name,
            dropped_datapoints=dropped_datapoints,
        )


@dataclass
class UploadDeployResult(DeployResult):
    uploaded: int = 0
    item_name: str = ""

    def __iadd__(self, other: "UploadDeployResult") -> "UploadDeployResult":
        if self.name != other.name:
            raise ValueError("Cannot add two DeployResult objects with different names")
        self.uploaded += other.uploaded

        if isinstance(other, DatapointDeployResult):
            return DatapointDeployResult(
                name=self.name, uploaded=self.uploaded, item_name=other.item_name, points=other.points
            )
        else:
            return self


@dataclass
class DatapointDeployResult(UploadDeployResult):
    points: int = 0

    def __iadd__(self, other: UploadDeployResult) -> UploadDeployResult:
        if self.name != other.name:
            raise ValueError("Cannot add two DeployResult objects with different names")
        super().__iadd__(other)
        if isinstance(other, DatapointDeployResult):
            self.points += other.points
        return self


class DeployResults(UserDict):
    def __init__(
        self,
        collection: Iterable[DeployResult],
        action: Literal["deploy", "clean", "purge", "pull"],
        dry_run: bool = False,
    ):
        super().__init__({entry.name: entry for entry in collection})
        self.action = action
        self.dry_run = dry_run

    @property
    def has_counts(self) -> bool:
        return any(isinstance(entry, ResourceDeployResult) for entry in self.data.values())

    @property
    def has_uploads(self) -> bool:
        return any(
            isinstance(entry, UploadDeployResult | ResourceContainerDeployResult) for entry in self.data.values()
        )

    def counts_table(
        self, exclude_columns: set[Literal["Created", "Deleted", "Changed", "Untouched", "Total"]] | None = None
    ) -> Table:
        table = Table(title=f"Summary of Resources {self.action.title()} operation:")
        prefix = "Would have " if self.dry_run else ""
        table.add_column("Resource", justify="right", width=30)
        if exclude_columns is None or "Created" not in exclude_columns:
            table.add_column(f"{prefix}Created", justify="right", style="green")
        if exclude_columns is None or "Deleted" not in exclude_columns:
            table.add_column(f"{prefix}Deleted", justify="right", style="red")
        if exclude_columns is None or "Changed" not in exclude_columns:
            table.add_column(f"{prefix}Changed", justify="right", style="magenta")
        if exclude_columns is None or "Untouched" not in exclude_columns:
            table.add_column("Untouched" if self.dry_run else "Unchanged", justify="right", style="cyan")
        if exclude_columns is None or "Total" not in exclude_columns:
            table.add_column("Total", justify="right")
        is_deploy = self.action == "deploy"
        for item in sorted(
            entry for entry in self.data.values() if entry is not None and isinstance(entry, ResourceDeployResult)
        ):
            if (
                item.created == 0
                and item.deleted == 0
                and item.changed == 0
                and item.unchanged == 0
                and item.total == 0
            ):
                continue

            row = [item.name]
            if exclude_columns is None or "Created" not in exclude_columns:
                row.append(f"{item.created:,}")
            if exclude_columns is None or "Deleted" not in exclude_columns:
                row.append(f"{item.deleted:,}")
            if exclude_columns is None or "Changed" not in exclude_columns:
                row.append(f"{item.changed:,}")
            if exclude_columns is None or "Untouched" not in exclude_columns:
                row.append(f"{item.unchanged:,}" if is_deploy else "-")
            if exclude_columns is None or "Total" not in exclude_columns:
                row.append(f"{item.total:,}")
            table.add_row(*row)

        return table

    def uploads_table(self) -> Table:
        table = Table(title=f"Summary of Data {self.action.title()} operation (data is always uploaded):")
        prefix = "Would have " if self.dry_run else ""
        table.add_column("Resource", justify="right")
        table.add_column(f"{prefix}Uploaded Data", justify="right", style="cyan")
        table.add_column("Item Type", justify="right")
        table.add_column("From files", justify="right", style="green")
        table.add_column(f"{prefix}Deleted Data", justify="right", style="red")
        for item in sorted(
            entry
            for entry in self.data.values()
            if isinstance(entry, UploadDeployResult | ResourceContainerDeployResult)
        ):
            if item.name == "raw.tables":
                # We skip this as we cannot count the number of datapoints in a raw table
                # and all we can do is to print a misleading 0 for deleted datapoints.
                continue

            if isinstance(item, UploadDeployResult):
                if isinstance(item, DatapointDeployResult):
                    datapoints = f"{item.points:,}"
                else:
                    datapoints = "-"
                table.add_row(item.name, datapoints, item.item_name, str(item.uploaded), "-")
            elif isinstance(item, ResourceContainerDeployResult):
                table.add_row(item.name, "-", item.item_name, "-", f"{item.dropped_datapoints:,}")

        return table

    def __getitem__(self, item: str) -> DeployResult:
        return self.data[item]

    def __setitem__(self, key: str, value: DeployResult) -> None:
        if key not in self.data:
            self.data[key] = value
        else:
            self.data[key] += value



from pydantic import Field, JsonValue

from cognite_toolkit._cdf_tk.constants import INSTANCE_EXTERNAL_ID_PATTERN, SPACE_FORMAT_PATTERN

from .base import BaseModelResource, ToolkitResource
from .container_field_definitions import ContainerReference
from .view_field_definitions import ViewReference


class NodeType(BaseModelResource):
    space: str = Field(
        description="The space where the instance is located.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the instance.",
        min_length=1,
        max_length=256,
        pattern=INSTANCE_EXTERNAL_ID_PATTERN,
    )


class InstanceSource(BaseModelResource):
    source: ViewReference | ContainerReference = Field(
        description="Reference to the view or container from where this source is inherited.",
    )
    properties: dict[str, JsonValue] | None = Field(
        default=None,
        description="Properties included in the source.",
    )


class NodeYAML(ToolkitResource):
    space: str = Field(
        description="The space where the node is located.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the node.",
        min_length=1,
        max_length=256,
        pattern=INSTANCE_EXTERNAL_ID_PATTERN,
    )
    type: NodeType | None = Field(
        default=None,
        description="The type of the node, defined by a direct relation to another node.",
    )
    sources: list[InstanceSource] | None = Field(
        default=None,
        description="List of sources (views) and their properties for this node.",
    )
    existing_version: int | None = Field(
        default=None,
        description="Fail the ingestion request if the node version is greater than this value.",
        ge=0,
    )


class EdgeYAML(ToolkitResource):
    space: str = Field(
        description="Id of the space that the edge belongs to.",
        min_length=1,
        max_length=43,
        pattern=SPACE_FORMAT_PATTERN,
    )
    external_id: str = Field(
        description="External-id of the edge.",
        min_length=1,
        max_length=256,
        pattern=INSTANCE_EXTERNAL_ID_PATTERN,
    )
    type: NodeType = Field(
        description="The type of the edge, defined by a direct relation to another edge.",
    )
    sources: list[InstanceSource] | None = Field(
        default=None,
        description="List of sources (views) and their properties for this edge.",
    )
    existing_version: int | None = Field(
        default=None,
        description="Version of the edge.",
        ge=0,
    )
    start_node: NodeType = Field(
        description="Reference to the node at the start of the edge.",
    )
    end_node: NodeType = Field(
        description="Reference to the node at the end of the edge.",
    )



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.resource_classes.robotics.location import RobotLocationYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.test_unit.utils import find_resources


def valid_robot_location_cases() -> Iterable:
    yield pytest.param(
        {
            "name": "Factory Floor A",
            "externalId": "factory_floor_a",
            "description": "Main production floor with assembly lines 1-5 and robot inspection stations",
        },
        id="full-location-with-description",
    )
    yield pytest.param(
        {
            "name": "Warehouse Section B",
            "externalId": "warehouse_b",
        },
        id="minimal-robot-location",
    )


def invalid_robot_location_cases() -> Iterable:
    yield pytest.param(
        {"name": "Missing External ID"},
        {"Missing required field: 'externalId'"},
        id="missing-required-field-external-id",
    )
    yield pytest.param(
        {
            "name": "Test Location",
            "externalId": "loc_001",
            "unknownField": "value",
        },
        {"Unused field: 'unknownField'"},
        id="unknown-field-present",
    )


class TestRobotLocationYAML:
    @pytest.mark.parametrize("data", list(find_resources("Location", "robotics")))
    def test_load_valid_robot_location_file(self, data: dict[str, object]) -> None:
        loaded = RobotLocationYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data", valid_robot_location_cases())
    def test_load_valid_robot_location(self, data: dict[str, object]) -> None:
        loaded = RobotLocationYAML.model_validate(data)
        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", invalid_robot_location_cases())
    def test_invalid_robot_location_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, RobotLocationYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)
        assert set(format_warning.errors) == expected_errors



from __future__ import annotations

import hashlib
import re
import zipfile
from pathlib import Path
from unittest.mock import MagicMock

import pytest
import requests
import yaml
from _pytest.monkeypatch import MonkeyPatch
from questionary import Choice

from cognite_toolkit._cdf_tk.commands.modules import ModulesCommand
from cognite_toolkit._cdf_tk.constants import MODULES
from cognite_toolkit._cdf_tk.data_classes import Package, Packages
from cognite_toolkit._cdf_tk.exceptions import ToolkitError
from cognite_toolkit._cdf_tk.tk_warnings.other import HighSeverityWarning
from tests.data import COMPLETE_ORG, EXTERNAL_PACKAGE
from tests.test_unit.utils import MockQuestionary

COMPLETE_ORG_MODULES = COMPLETE_ORG / MODULES


@pytest.fixture(scope="session")
def selected_packages() -> Packages:
    return Packages.load(COMPLETE_ORG_MODULES)


@pytest.fixture(scope="session")
def selected_packages_location() -> Path:
    return COMPLETE_ORG_MODULES


class MockResponse:
    def __init__(self, content, status_code=200, headers=None):
        self._content = content
        self.status_code = status_code
        self.headers = headers if headers is not None else {"content-length": str(len(content))}
        self.raise_for_status_called = False

    def iter_content(self, chunk_size=8192):
        yield self._content

    def raise_for_status(self):
        self.raise_for_status_called = True
        if self.status_code >= 400:
            raise requests.exceptions.HTTPError(f"HTTP Error: {self.status_code}")


class TestModulesCommand:
    def test_modules_command(
        self, selected_packages: Packages, selected_packages_location: Path, tmp_path: Path
    ) -> None:
        assert selected_packages is not None

        target_path = tmp_path / "repo_root"

        cmd = ModulesCommand(print_warning=True, skip_tracking=True)
        cmd._create(
            organization_dir=target_path,
            selected_packages=selected_packages,
            environments=["dev"],
            mode=None,
            modules_source_path=selected_packages_location,
        )

        assert Path(target_path).exists()
        assert Path(target_path / "modules" / "my_example_module").exists()

    def test_modules_command_with_env(
        self, selected_packages: Packages, selected_packages_location: Path, tmp_path: Path
    ) -> None:
        assert selected_packages is not None

        target_path = tmp_path / "repo_root"

        cmd = ModulesCommand(print_warning=True, skip_tracking=True)
        cmd._create(
            organization_dir=target_path,
            selected_packages=selected_packages,
            environments=["dev", "prod"],
            mode=None,
            modules_source_path=selected_packages_location,
        )

        assert Path(target_path / "config.dev.yaml").exists()
        assert Path(target_path / "config.prod.yaml").exists()

    def test_config(self, selected_packages: Packages, selected_packages_location: Path, tmp_path: Path) -> None:
        assert selected_packages is not None

        target_path = tmp_path / "repo_root"

        cmd = ModulesCommand(print_warning=True, skip_tracking=True)
        cmd._create(
            organization_dir=target_path,
            selected_packages=selected_packages,
            environments=["dev"],
            mode=None,
            modules_source_path=selected_packages_location,
        )

        config = yaml.safe_load(Path(target_path / "config.dev.yaml").read_text())
        assert config["variables"]["modules"]["my_example_module"]["var"] == "one"

    def test_config_external_modules(self, tmp_path: Path) -> None:
        target_path = tmp_path / "repo_root"

        selected_packages = Packages.load(EXTERNAL_PACKAGE)
        selected_packages_location = EXTERNAL_PACKAGE

        cmd = ModulesCommand(print_warning=True, skip_tracking=True)
        cmd._create(
            organization_dir=target_path,
            selected_packages=selected_packages,
            environments=["dev"],
            mode=None,
            modules_source_path=selected_packages_location,
        )

        config = yaml.safe_load(Path(target_path / "config.dev.yaml").read_text())
        assert config["variables"]["modules"]["shared_var"] == "shared"
        assert config["variables"]["modules"]["external_module_1"]["var"] == "one"
        assert config["variables"]["modules"]["external_module_2"]["var"] == "two"

    def test_adding(
        self, selected_packages: Packages, selected_packages_location: Path, tmp_path: Path, monkeypatch: MonkeyPatch
    ) -> None:
        target_path = tmp_path / "repo_root"
        cmd = ModulesCommand(print_warning=True, skip_tracking=True, module_source_dir=COMPLETE_ORG_MODULES)

        first_batch = Packages({"small_1": selected_packages["small_1"]})
        second_batch = Packages({"small_2": selected_packages["small_2"]})

        cmd._create(
            organization_dir=target_path,
            selected_packages=first_batch,
            environments=["qa"],
            mode=None,
            modules_source_path=selected_packages_location,
        )
        with monkeypatch.context() as m:
            # Mocking questionary such that questionary.confirm.ask() returns True.
            questionary_mock = MagicMock()
            # MagicMock will always return other MagicMock objects
            # which when evaluated will return True.

            m.setattr("cognite_toolkit._cdf_tk.commands.modules.questionary", questionary_mock)
            cmd._create(
                organization_dir=target_path,
                selected_packages=second_batch,
                environments=["qa"],
                mode="update",
                modules_source_path=selected_packages_location,
            )

        config = yaml.safe_load(Path(target_path / "config.qa.yaml").read_text())
        assert config["variables"]["modules"]["my_example_module"]["var"] is not None
        assert (target_path / "modules" / "my_example_module").is_dir()

        assert config["variables"]["modules"]["my_file_expand_module"]["var"] is not None
        assert (target_path / "modules" / "my_file_expand_module").is_dir()

    def test_add_without_config_yaml(
        self, tmp_path: Path, monkeypatch: MonkeyPatch, modules_command_with_cached_download
    ) -> None:
        cmd = modules_command_with_cached_download(
            print_warning=True, skip_tracking=True, module_source_dir=COMPLETE_ORG_MODULES
        )
        dummy_resource = "space: my_space"
        my_org = tmp_path / "my_org"
        modules = my_org / "modules"
        filepath = modules / "my_module" / "data_models" / "my.Space.yaml"
        filepath.parent.mkdir(parents=True, exist_ok=True)
        filepath.write_text(dummy_resource)

        def select_package(choices: list[Choice]) -> Package:
            selected_package = next((c for c in choices if "complete organization" in c.title.lower()), None)
            assert selected_package is not None
            return selected_package.value

        def select_first_module(choices: list[Choice]) -> list:
            selected_module = next((c for c in choices if "my_example_module" == c.title), None)
            assert selected_module is not None
            return [selected_module.value]

        answers = [select_package, select_first_module, False, False]

        with MockQuestionary(ModulesCommand.__module__, monkeypatch, answers):
            cmd.add(my_org)

        yaml_file_count = len(list(modules.rglob("*.yaml")))

        assert yaml_file_count > 1, "Expected new yaml files to b created"

        def select_second_module(choices: list[Choice]) -> list:
            selected_module = next((c for c in choices if "my_file_expand_module" == c.title), None)
            assert selected_module is not None
            return [selected_module.value]

        answers = [select_package, select_second_module, False, False]

        with MockQuestionary(ModulesCommand.__module__, monkeypatch, answers):
            cmd.add(my_org)

        new_yaml_file_count = len(list(modules.rglob("*.yaml")))

        assert new_yaml_file_count > yaml_file_count, "Expected new yaml files to be created"

    def test_context_manager_scope(self):
        with ModulesCommand(module_source_dir=COMPLETE_ORG_MODULES) as cmd:
            first = Path(cmd._temp_download_dir / "test.txt")
            first.write_text("This is a test file.")
            assert first.exists()
        assert not first.exists(), "File should not exist after context manager exits"

    @pytest.fixture(autouse=True)
    def reset_cdf_toml_singleton(self):
        global _CDF_TOML
        _CDF_TOML = None
        yield
        _CDF_TOML = None

    def test_download_success(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        dummy_file_content = b"PK\x05\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"

        mock_response = MockResponse(dummy_file_content, status_code=200)
        monkeypatch.setattr(requests, "get", MagicMock(return_value=mock_response))

        cmd = ModulesCommand(print_warning=True, skip_tracking=True, module_source_dir=COMPLETE_ORG_MODULES)
        output_zip_path = tmp_path / "test_file.zip"

        cmd._download(url="http://example.com/test.zip", file_path=output_zip_path)

        requests.get.assert_called_once_with("http://example.com/test.zip", stream=True)
        assert output_zip_path.exists()
        assert output_zip_path.read_bytes() == dummy_file_content

    def test_download_errors_http_error(
        self,
        tmp_path: Path,
        monkeypatch: MonkeyPatch,
    ) -> None:
        test_url = "http://example.com/test_file.zip"
        output_path = tmp_path / "test_file.zip"

        # Arrange: Mock requests.get to return a MockResponse with a 404 status
        monkeypatch.setattr(requests, "get", MagicMock(return_value=MockResponse(b"", status_code=404)))

        # Act & Assert
        with pytest.raises(ToolkitError) as excinfo:
            ModulesCommand(module_source_dir=COMPLETE_ORG_MODULES)._download(test_url, output_path)

        assert isinstance(excinfo.value.__cause__, requests.exceptions.HTTPError)

    def test_download_errors_request_exception(
        self,
        tmp_path: Path,
        monkeypatch: MonkeyPatch,
    ) -> None:
        test_url = "http://example.com/test_file.zip"
        output_path = tmp_path / "test_file.zip"

        # Arrange: Mock requests.get to raise a RequestException directly
        monkeypatch.setattr(
            requests, "get", MagicMock(side_effect=requests.exceptions.RequestException("Connection aborted."))
        )

        # Act & Assert
        with pytest.raises(ToolkitError) as excinfo:
            ModulesCommand(module_source_dir=COMPLETE_ORG_MODULES)._download(test_url, output_path)

        assert "Error downloading file" in str(excinfo.value)
        assert isinstance(excinfo.value.__cause__, requests.exceptions.RequestException)
        assert "Connection aborted." in str(excinfo.value.__cause__)

    def test_unpack_errors_bad_zip_file(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        url_suffix = "corrupt_file.zip"
        output_path = tmp_path / url_suffix

        output_path.touch()

        mock_zipfile_instance = MagicMock()
        mock_zipfile_instance.__enter__.side_effect = zipfile.BadZipFile("File is not a zip file")

        monkeypatch.setattr(zipfile, "ZipFile", MagicMock(return_value=mock_zipfile_instance))

        with pytest.raises(ToolkitError) as excinfo:
            ModulesCommand(module_source_dir=COMPLETE_ORG_MODULES)._unpack(output_path)

        assert isinstance(excinfo.value.__cause__, zipfile.BadZipFile)
        assert "File is not a zip file" in str(excinfo.value.__cause__)

    def test_unpack_errors_os_error_during_write(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        url_suffix = "valid_archive.zip"
        output_path = tmp_path / url_suffix

        with zipfile.ZipFile(output_path, "w") as zf:
            zf.writestr("dummy_file.txt", "content")

        mock_zipfile_ref = MagicMock()
        mock_zipfile_ref.extractall.side_effect = OSError("No space left on device")
        mock_zipfile_ref.__enter__.return_value = mock_zipfile_ref
        mock_zipfile_ref.__exit__.return_value = None

        monkeypatch.setattr(zipfile, "ZipFile", MagicMock(return_value=mock_zipfile_ref))

        with pytest.raises(ToolkitError) as excinfo:
            ModulesCommand(module_source_dir=COMPLETE_ORG_MODULES)._unpack(output_path)

        assert isinstance(excinfo.value.__cause__, OSError)
        assert "No space left on device" in str(excinfo.value.__cause__)

    def test_checksum_format(self, tmp_path: Path) -> None:
        invalid_checksum = "1234567890abcdef1234567890abcdef1234567890abcdef1234567890abcdef"
        with pytest.raises(ToolkitError) as excinfo:
            ModulesCommand(module_source_dir=COMPLETE_ORG_MODULES)._validate_checksum(
                invalid_checksum, Path(tmp_path / "test_file.zip")
            )

        assert "Unsupported checksum format" in str(excinfo.value)

    def test_checksum_success(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        file_path = tmp_path / "test_file.zip"
        dummy_file_content = b"PK\x05\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        file_path.write_bytes(dummy_file_content)

        expected_checksum = f"sha256:{hashlib.sha256(dummy_file_content).hexdigest()}"

        cmd = ModulesCommand(print_warning=True, skip_tracking=True, module_source_dir=COMPLETE_ORG / MODULES)
        try:
            cmd._validate_checksum(
                checksum=expected_checksum,
                file_path=file_path,  # Pass the correct Path object
            )
        except ToolkitError as e:
            pytest.fail(f"'_validate_checksum' raised an unexpected ToolkitError: {e}")

    def test_download_deletes_existing_file(self, tmp_path: Path, monkeypatch: MonkeyPatch) -> None:
        """Test that _download method deletes existing zip files before downloading."""
        # Create a stale zip file that should be deleted
        stale_file_path = tmp_path / "test_file.zip"
        stale_content = b"stale content"
        stale_file_path.write_bytes(stale_content)

        # Verify the stale file exists
        assert stale_file_path.exists()
        assert stale_file_path.read_bytes() == stale_content

        # Mock the HTTP response with new content
        new_content = b"new content from download"
        mock_response = MockResponse(new_content, status_code=200)
        monkeypatch.setattr(requests, "get", MagicMock(return_value=mock_response))

        cmd = ModulesCommand(print_warning=True, skip_tracking=True, module_source_dir=COMPLETE_ORG / MODULES)

        # Call _download - this should delete the existing file and download new content
        cmd._download(url="http://example.com/test.zip", file_path=stale_file_path)

        # Verify the file was deleted and replaced with new content
        assert stale_file_path.exists()
        assert stale_file_path.read_bytes() == new_content
        assert stale_file_path.read_bytes() != stale_content

        # Verify the HTTP request was made
        requests.get.assert_called_once_with("http://example.com/test.zip", stream=True)

    def test_iterate_modules_finds_modules_in_temp_download_dir(self, tmp_path: Path) -> None:
        """Test that iterate_modules can find modules in the _temp_download_dir.

        This test verifies that when modules are downloaded to the temporary directory,
        the iterate_modules function can properly discover and iterate over them.
        The test creates a mock module structure with the required resource directories
        (like 'data_models') that the module discovery logic recognizes.
        """
        from cognite_toolkit._cdf_tk.utils.modules import iterate_modules

        cmd = ModulesCommand(print_warning=True, skip_tracking=True, module_source_dir=COMPLETE_ORG / MODULES)

        # Create a mock module structure in the temp download directory
        # This simulates what would happen when modules are downloaded
        mock_module_dir = cmd._temp_download_dir / "test_module"
        mock_module_dir.mkdir(parents=True, exist_ok=True)

        # Create a resource directory (e.g., 'data_models') that LOADER_BY_FOLDER_NAME recognizes
        # This is required for the module to be identified as a valid module
        resource_dir = mock_module_dir / "data_models"
        resource_dir.mkdir()

        # Create a sample file in the resource directory
        sample_file = resource_dir / "sample.yaml"
        sample_file.write_text("test content")

        # Now test that iterate_modules can find this module
        modules_found = list(iterate_modules(cmd._temp_download_dir))

        # Should find at least one module
        assert len(modules_found) > 0, f"Expected to find modules in {cmd._temp_download_dir}"

        # Verify the module structure
        module_dir, files = modules_found[0]
        assert module_dir == mock_module_dir
        assert len(files) > 0
        assert any(file.name == "sample.yaml" for file in files)

        # Clean up
        import shutil

        shutil.rmtree(mock_module_dir)

    def test_checksum_mismatch_prints_warning(self, tmp_path: Path, capsys) -> None:
        file_path = tmp_path / "mismatch.zip"
        # Write some bytes so we get a deterministic SHA256
        file_bytes = b"dummy-bytes-for-checksum-test"
        file_path.write_bytes(file_bytes)

        # Intentionally use a different checksum than the file's actual hash
        wrong_checksum = "sha256:" + hashlib.sha256(b"some-other-content").hexdigest()

        cmd = ModulesCommand(print_warning=True, skip_tracking=True, module_source_dir=COMPLETE_ORG / MODULES)
        cmd._validate_checksum(wrong_checksum, file_path)

        assert len(cmd.warning_list) == 1
        warning = cmd.warning_list[0]
        assert isinstance(warning, HighSeverityWarning)
        # Expect: two SHA256 hex hashes in the message, one for provided and one for calculated
        pattern = r"^The provided checksum sha256:[0-9a-f]{64} does not match downloaded file hash sha256:[0-9a-f]{64}"
        assert re.search(pattern, warning.message_raw)



from __future__ import annotations

import hashlib
import io
import itertools
import json as JSON
import zipfile
from collections import defaultdict
from collections.abc import Callable, Iterable, Sequence
from pathlib import Path
from typing import Any, BinaryIO, TextIO, cast
from unittest.mock import MagicMock

import pandas as pd
from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client._api.iam import IAMAPI
from cognite.client.credentials import OAuthClientCredentials
from cognite.client.data_classes import (
    Database,
    DataSet,
    ExtractionPipeline,
    ExtractionPipelineConfig,
    ExtractionPipelineConfigWrite,
    FileMetadata,
    FunctionCall,
    Group,
    GroupList,
    Table,
    TableList,
    ThreeDModel,
    WorkflowVersion,
    capabilities,
)
from cognite.client.data_classes._base import CogniteResource, T_CogniteResource
from cognite.client.data_classes.capabilities import AllProjectsScope, ProjectCapability, ProjectCapabilityList
from cognite.client.data_classes.data_modeling import (
    DataModelList,
    Edge,
    EdgeApply,
    EdgeApplyResult,
    EdgeApplyResultList,
    EdgeId,
    EdgeList,
    InstancesApplyResult,
    InstancesDeleteResult,
    InstancesResult,
    Node,
    NodeApply,
    NodeApplyResult,
    NodeApplyResultList,
    NodeId,
    Space,
    VersionedDataModelingId,
    View,
)
from cognite.client.data_classes.data_modeling.graphql import DMLApplyResult
from cognite.client.data_classes.data_modeling.ids import DataModelIdentifier, InstanceId
from cognite.client.data_classes.functions import FunctionsStatus
from cognite.client.data_classes.iam import CreatedSession, GroupWrite, ProjectSpec, TokenInspection
from cognite.client.utils._text import to_camel_case
from cognite.client.utils.useful_types import SequenceNotStr
from requests import Response

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.graphql_data_models import GraphQLDataModelWrite
from cognite_toolkit._cdf_tk.client.data_classes.project import ProjectStatus, ProjectStatusList
from cognite_toolkit._cdf_tk.client.data_classes.raw import RawDatabase
from cognite_toolkit._cdf_tk.client.testing import ToolkitClientMock
from cognite_toolkit._cdf_tk.constants import INDEX_PATTERN
from cognite_toolkit._cdf_tk.cruds import FileCRUD
from cognite_toolkit._cdf_tk.utils import calculate_hash
from cognite_toolkit._cdf_tk.utils.auth import CLIENT_NAME

from .config import API_RESOURCES
from .data_classes import APIResource, AuthGroupCalls

TEST_FOLDER = Path(__file__).resolve().parent.parent

_ALL_CAPABILITIES = []
for cap, (scopes, names) in capabilities._VALID_SCOPES_BY_CAPABILITY.items():
    for action, scope in itertools.product(cap.Action, scopes):
        try:
            _ALL_CAPABILITIES.append(cap([action], scope=scope()))
        except TypeError:
            pass
del cap, scopes, names, action, scope


class LookUpAPIMock:
    def __init__(self, allow_reverse_lookup: bool = False):
        self._reverse_cache: dict[int, str] = {}
        self._allow_reverse_lookup = allow_reverse_lookup

    @staticmethod
    def create_id(string: str, allow_empty: bool = False) -> int:
        if allow_empty and string == "":
            return 0
        # This simulates CDF setting the internal ID.
        # By using hashing, we will always get the same ID for the same string.
        # Thus, the ID will be consistent between runs and can be used in snapshots.
        hash_object = hashlib.sha256(string.encode())
        hex_dig = hash_object.hexdigest()
        hash_int = int(hex_dig[:16], 16)
        return hash_int

    def id(
        self, external_id: str | SequenceNotStr[str], is_dry_run: bool = False, allow_empty: bool = False
    ) -> int | list[int]:
        if isinstance(external_id, str):
            id_ = self.create_id(external_id, allow_empty)
            if id_ not in self._reverse_cache:
                self._reverse_cache[id_] = external_id
            return id_
        output: list[int] = []
        for ext_id in external_id:
            id_ = self.create_id(ext_id, allow_empty)
            if id_ not in self._reverse_cache:
                self._reverse_cache[id_] = ext_id
            output.append(id_)
        return output

    @staticmethod
    def create_external_id(id: int | Sequence[int]) -> str | list[str]:
        """Creates an external ID from an ID."""
        if isinstance(id, int):
            return f"external_{id}"
        return [f"external_{i}" for i in id]

    def external_id(
        self,
        id: int | Sequence[int],
    ) -> str | list[str]:
        try:
            return self._reverse_cache[id] if isinstance(id, int) else [self._reverse_cache[i] for i in id]
        except KeyError:
            if self._allow_reverse_lookup:
                return self.create_external_id(id) if isinstance(id, int) else [self.create_external_id(i) for i in id]
            else:
                raise RuntimeError(f"{type(self).__name__} does not support reverse lookup before lookup")


class ApprovalToolkitClient:
    """A mock CogniteClient that is used for testing the clean, deploy commands
    of the cognite-toolkit.

    Args:
        mock_client: The mock client to use.

    """

    def __init__(self, mock_client: ToolkitClientMock, allow_reverse_lookup: bool = False):
        self._return_verify_resources = False
        self.mock_client = mock_client
        credentials = MagicMock(spec=OAuthClientCredentials)
        credentials.client_id = "toolkit-client-id"
        credentials.client_secret = "toolkit-client-secret"
        credentials.token_url = "https://toolkit.auth.com/oauth/token"
        credentials.scopes = ["ttps://pytest-field.cognitedata.com/.default"]
        project = "test_project"
        self.mock_client.config = ToolkitClientConfig(
            client_name=CLIENT_NAME,
            project=project,
            credentials=credentials,
            is_strict_validation=False,
        )
        # This is used to simulate the existing resources in CDF
        self._existing_resources: dict[str, list[CogniteResource]] = defaultdict(list)
        # This is used to log all delete operations
        self._deleted_resources: dict[str, list[str | int | dict[str, Any]]] = defaultdict(list)
        # This is used to log all create operations
        self.created_resources: dict[str, list[CogniteResource | dict[str, Any]]] = defaultdict(list)

        # This is used to log all operations
        self._delete_methods: dict[str, list[MagicMock]] = defaultdict(list)
        self._create_methods: dict[str, list[MagicMock]] = defaultdict(list)
        self._retrieve_methods: dict[str, list[MagicMock]] = defaultdict(list)
        self._inspect_methods: dict[str, list[MagicMock]] = defaultdict(list)
        self._post_methods: dict[str, list[MagicMock]] = defaultdict(list)

        # Set the side effect of the MagicMock to the real method
        self.mock_client.iam.compare_capabilities.side_effect = IAMAPI.compare_capabilities
        self.mock_client.iam.sessions.create.return_value = CreatedSession(
            id=1234, status="READY", nonce="123", type="CLIENT_CREDENTIALS", client_id="12345-12345-12345-12345"
        )
        # Set functions to be activated
        self.mock_client.functions.status.return_value = FunctionsStatus(status="activated")

        # Use Hybrid project
        self.mock_client.project.status.return_value = ProjectStatusList(
            [ProjectStatus(url_name=project, data_modeling_status="HYBRID")], cognite_client=mock_client
        )

        # Activate authorization_header()
        self.mock_client.config.credentials.authorization_header.return_value = ("Bearer", "123")
        # Set project
        self.mock_client.config.project = project
        self.mock_client.config.base_url = "https://bluefield.cognitedata.com"
        # Setup mock for all lookup methods
        for method_name, lookup_api in self.mock_client.lookup.__dict__.items():
            if method_name.startswith("_") or method_name == "method_calls":
                continue
            mock_lookup = LookUpAPIMock(allow_reverse_lookup)
            lookup_api.id.side_effect = mock_lookup.id
            lookup_api.external_id.side_effect = mock_lookup.external_id
        self.mock_client.verify.authorization.return_value = []

        # Setup all mock methods
        for resource in API_RESOURCES:
            parts = resource.api_name.split(".")
            mock_api = mock_client
            for part in parts:
                if not hasattr(mock_api, part):
                    raise ValueError(f"Invalid api name {resource.api_name}, could not find {part}")
                # To avoid registering the side effect on the mock_client.post.post and use
                # just mock_client.post instead, we need to skip the "step into" post mock here.
                if part != "post":
                    mock_api = getattr(mock_api, part)
            for method_type, methods in resource.methods.items():
                method_factory: Callable = {
                    "create": self._create_create_method,
                    "delete": self._create_delete_method,
                    "retrieve": self._create_retrieve_method,
                    "inspect": self._create_inspect_method,
                    "post": self._create_post_method,
                    "upsert": self._create_retrieve_method,
                    "update": self._create_post_method,
                }[method_type]
                method_dict = {
                    "create": self._create_methods,
                    "delete": self._delete_methods,
                    "retrieve": self._retrieve_methods,
                    "inspect": self._inspect_methods,
                    "post": self._post_methods,
                    "upsert": self._retrieve_methods,
                    "update": self._post_methods,
                }[method_type]
                for mock_method in methods:
                    if not hasattr(mock_api, mock_method.api_class_method):
                        raise ValueError(
                            f"Invalid api method {mock_method.api_class_method} for resource {resource.api_name}"
                        )
                    method = getattr(mock_api, mock_method.api_class_method)
                    method.side_effect = method_factory(resource, mock_method.mock_class_method, mock_client)
                    method_dict[resource.resource_cls.__name__].append(method)

    @property
    def client(self) -> ToolkitClient:
        """Returns a mock ToolkitClient"""
        return cast(ToolkitClient, self.mock_client)

    @property
    def return_verify_resources(self) -> bool:
        return self._return_verify_resources

    @return_verify_resources.setter
    def return_verify_resources(self, value: bool) -> None:
        """This is used to return the resource that are used for verication.

        Caveat: This only applies to Spaces, DataSets, and ExtractionPipeline.

        The use case is that these are used in verification of other resources.
        """
        self._return_verify_resources = value

    def append(self, resource_cls: type[CogniteResource], items: CogniteResource | Sequence[CogniteResource]) -> None:
        """This is used to simulate existing resources in CDF.

        Args:
            resource_cls: The type of resource this is.
            items: The list of resources to append.

        """
        if isinstance(items, Sequence):
            self._existing_resources[resource_cls.__name__].extend(items)
        else:
            self._existing_resources[resource_cls.__name__].append(items)

    def clear_cdf_resources(self, resource_cls: type[CogniteResource]) -> None:
        """Clears the existing resources in CDF.

        Args:
            resource_cls: The type of resource to clear.

        """
        self._existing_resources[resource_cls.__name__].clear()

    def _create_delete_method(self, resource: APIResource, mock_method: str, client: CogniteClient) -> Callable:
        deleted_resources = self._deleted_resources
        resource_cls = resource.resource_cls

        def delete_id_external_id(
            id: int | Sequence[int] | None = None,
            external_id: str | Sequence[str] | None = None,
            **_,
        ) -> list:
            deleted = []
            if not isinstance(id, str) and isinstance(id, Sequence):
                deleted.extend({"id": i} for i in id)
            elif isinstance(id, int):
                deleted.append({"id": id})
            if isinstance(external_id, str):
                deleted.append({"externalId": external_id})
            elif isinstance(external_id, Sequence):
                deleted.extend({"externalId": i} for i in external_id)
            if deleted:
                deleted_resources[resource_cls.__name__].extend(deleted)
            return deleted

        def delete_data_modeling(ids: VersionedDataModelingId | Sequence[VersionedDataModelingId]) -> list:
            deleted = []
            if isinstance(ids, VersionedDataModelingId | InstanceId):
                deleted.append(ids.dump(camel_case=True))
            elif isinstance(ids, Sequence):
                deleted.extend([id.dump(camel_case=True) for id in ids])
            if deleted:
                deleted_resources[resource_cls.__name__].extend(deleted)
            return deleted

        def delete_instances(
            nodes: NodeId | Sequence[NodeId] | tuple[str, str] | Sequence[tuple[str, str]] | None = None,
            edges: EdgeId | Sequence[EdgeId] | tuple[str, str] | Sequence[tuple[str, str]] | None = None,
        ) -> InstancesDeleteResult:
            deleted = []
            if isinstance(nodes, NodeId):
                deleted.append(nodes.dump(camel_case=True, include_instance_type=True))
            elif isinstance(nodes, tuple):
                deleted.append(NodeId(*nodes).dump(camel_case=True, include_instance_type=True))
            elif isinstance(edges, EdgeId):
                deleted.append(edges.dump(camel_case=True, include_instance_type=True))
            elif isinstance(edges, tuple):
                deleted.append(EdgeId(*edges).dump(camel_case=True, include_instance_type=True))
            elif isinstance(nodes, Sequence):
                deleted.extend(
                    [
                        node.dump(camel_case=True, include_instance_type=True) if isinstance(node, NodeId) else node
                        for node in nodes
                    ]
                )
            elif isinstance(edges, Sequence):
                deleted.extend(
                    [
                        edge.dump(camel_case=True, include_instance_type=True) if isinstance(edge, EdgeId) else edge
                        for edge in edges
                    ]
                )

            if deleted:
                deleted_resources[resource_cls.__name__].extend(deleted)

            if nodes:
                return InstancesDeleteResult(nodes=deleted, edges=[])
            elif edges:
                return InstancesDeleteResult(nodes=[], edges=deleted)
            else:
                return InstancesDeleteResult(nodes=[], edges=[])

        def delete_space(spaces: str | Sequence[str]) -> list:
            deleted = []
            if isinstance(spaces, str):
                deleted.append(spaces)
            elif isinstance(spaces, Sequence):
                deleted.extend(spaces)
            if deleted:
                deleted_resources[resource_cls.__name__].extend(deleted)
            return deleted

        def delete_raw(db_name: str | Sequence[str], name: str | Sequence[str] | None = None) -> list:
            if name:
                deleted = [{"db_name": db_name, "name": name if isinstance(name, str) else sorted(name)}]
            else:
                deleted = [{"db_name": name} for name in (db_name if isinstance(db_name, Sequence) else [db_name])]
            deleted_resources[resource_cls.__name__].extend(deleted)
            return deleted

        available_delete_methods = {
            fn.__name__: fn
            for fn in [
                delete_id_external_id,
                delete_instances,
                delete_raw,
                delete_data_modeling,
                delete_space,
            ]
        }
        if mock_method not in available_delete_methods:
            raise ValueError(
                f"Invalid mock delete method {mock_method} for resource {resource_cls.__name__}. "
                f"Supported {list(available_delete_methods)}"
            )

        method = available_delete_methods[mock_method]
        return method

    def _create_create_method(self, resource: APIResource, mock_method: str, client: CogniteClient) -> Callable:
        created_resources = self.created_resources
        write_resource_cls = resource.write_cls
        write_list_cls = resource.write_list_cls
        resource_cls = resource.resource_cls
        resource_list_cls = resource.list_cls

        def _create(*args, **kwargs) -> Sequence:
            created = []
            is_single_resource: bool | None = None
            for value in itertools.chain(args, kwargs.values()):
                if isinstance(value, Iterable):
                    values = value
                else:
                    values = [value]
                for item in values:
                    if isinstance(item, write_resource_cls):
                        created.append(item)
                        if is_single_resource is None:
                            is_single_resource = True
                        else:
                            is_single_resource = False
                    elif isinstance(item, Sequence) and all(isinstance(v, write_resource_cls) for v in item):
                        created.extend(item)
                        is_single_resource = False
                    elif isinstance(item, str) and issubclass(write_resource_cls, RawDatabase):
                        created.append(Database(name=item))
            created_resources[resource_cls.__name__].extend(created)
            if resource_cls is View:
                return write_list_cls(created)
            if resource_list_cls is GroupList:
                # Groups needs special handling to convert the write to read
                # to account for Unknown ACLs.
                return resource_list_cls(_group_write_to_read(c) for c in created)

            read_list = resource_list_cls.load(
                [
                    {
                        # These are server set fields, so we need to set them manually
                        # Note that many of them are only used for certain resources. This is not a problem
                        # as any extra fields are ignored by the load method.
                        "isGlobal": False,  # Data Modeling
                        "lastUpdatedTime": 0,
                        "createdTime": 0,
                        "writable": True,  # Data Modeling
                        "ignoreNullFields": False,  # Transformations
                        "usedFor": "nodes",  # Views
                        "timeSeriesCount": 10,  # Datapoint subscription
                        "updatedTime": 0,  # Robotics
                        "id": 42,  # LocationFilters
                        "status": "connected",  # Hosted Extractor Job
                        "targetStatus": "paused",  # Hosted Extractor Job
                        **c.dump(camel_case=True),
                    }
                    for c in created
                ],
                cognite_client=client,
            )
            return read_list

        def create_multiple(*args, **kwargs) -> Any:
            return _create(*args, **kwargs)

        def create_single(*args, **kwargs) -> Sequence:
            return _create(*args, **kwargs)[0]

        def create_filemetadata(*args, **kwargs) -> tuple[FileMetadata, str]:
            return create_single(*args, **kwargs), "upload link"

        def create_raw_table(db_name: str, name: str | list[str]) -> Table | TableList:
            if isinstance(name, str):
                created = Table(name=name, created_time=1)
                created_resources[resource_cls.__name__].append(created)
                return created
            else:
                created_list = TableList([Table(name=n, created_time=1) for n in name])
                created_resources[resource_cls.__name__].extend(created_list)
                return created_list

        def _group_write_to_read(group: GroupWrite) -> Group:
            return Group(
                name=group.name,
                source_id=group.source_id,
                capabilities=group.capabilities,
                metadata=group.metadata,
                members=group.members,
            )

        def upsert(*args, **kwargs) -> Any:
            upserted = []
            for value in itertools.chain(args, kwargs.values()):
                if isinstance(value, write_resource_cls):
                    upserted.append(value)
                elif isinstance(value, Sequence) and all(isinstance(v, write_resource_cls) for v in value):
                    upserted.extend(value)
            created_resources[resource_cls.__name__].extend(upserted)

            read_resource_objects: list[dict[str, object]] = [
                {
                    "isGlobal": False,
                    "lastUpdatedTime": 0,
                    "createdTime": 0,
                    **c.dump(camel_case=True),
                }
                for c in upserted
            ]
            if resource_cls is WorkflowVersion:
                for item in read_resource_objects:
                    if "workflowDefinition" not in item:
                        item["workflowDefinition"] = {}
                    if not isinstance(item["workflowDefinition"], dict):
                        raise TypeError(
                            f"Expected 'workflowDefinition' to be a dict, got {type(item['workflowDefinition'])}"
                        )
                    item["workflowDefinition"]["hash"] = "123"

            return resource_list_cls.load(read_resource_objects, cognite_client=client)

        def _create_dataframe_info(dataframe: pd.DataFrame) -> dict[str, Any]:
            return {
                "shape": "x".join(map(str, dataframe.shape)),
                "nan_count": int(dataframe.isna().sum().sum()),
                "null_count": int(dataframe.isnull().sum().sum()),
                "empty_count": int(dataframe[dataframe == ""].count().sum()),
                # We round float to 4 decimals places to avoid issues with floating point precision on different systems
                # as this is stored a snapshot.
                "first_row": dataframe.iloc[0].apply(lambda x: round(x, 4) if isinstance(x, float) else x).to_dict(),
                "last_row": dataframe.iloc[-1].apply(lambda x: round(x, 4) if isinstance(x, float) else x).to_dict(),
                "index_name": dataframe.index.name if dataframe.index.name else "missing",
            }

        def insert_dataframe(*args, **kwargs) -> None:
            args = list(args)
            kwargs = dict(kwargs)
            dataframe_info: dict[str, Any] = {}
            for arg in list(args):
                if isinstance(arg, pd.DataFrame):
                    args.remove(arg)
                    dataframe_info = _create_dataframe_info(arg)
                    break
            for key in list(kwargs):
                if isinstance(kwargs[key], pd.DataFrame):
                    value = kwargs.pop(key)
                    dataframe_info = _create_dataframe_info(value)
                    break
            if not dataframe_info:
                raise ValueError("No dataframe found in arguments")
            name = "_".join([str(arg) for arg in itertools.chain(args, kwargs.values())])
            if not name:
                name = "missing"
            created_resources[resource_cls.__name__].append(
                {"name": name, "args": args, "kwargs": kwargs, "dataframe": dataframe_info}
            )

        def upload(*args, **kwargs) -> None:
            name = ""
            for k, v in kwargs.items():
                if isinstance(v, Path) or (isinstance(v, str) and Path(v).exists()):
                    # The index pattern is used to ensure unique names. This index
                    # is removed as we do not care whether the order of the files are uploaded
                    filepath = Path(v)
                    filepath = filepath.with_name(INDEX_PATTERN.sub("", filepath.name))

                    try:
                        kwargs[k] = "/".join(filepath.relative_to(TEST_FOLDER).parts)
                    except ValueError:
                        kwargs[k] = "/".join(filepath.parts)
                    name = filepath.name

            created_resources[resource_cls.__name__].append(
                {
                    "name": name,
                    "args": list(args),
                    "kwargs": dict(kwargs),
                }
            )

        def create_instances(
            nodes: NodeApply | Sequence[NodeApply] | None = None,
            edges: EdgeApply | Sequence[EdgeApply] | None = None,
            **kwargs,
        ) -> InstancesApplyResult:
            created_nodes = []
            if isinstance(nodes, NodeApply):
                created_nodes.append(nodes)
            elif isinstance(nodes, Sequence) and all(isinstance(v, NodeApply) for v in nodes):
                created_nodes.extend(nodes)

            created_edges = []
            if isinstance(edges, EdgeApply):
                created_edges.append(edges)
            elif isinstance(edges, Sequence) and all(isinstance(v, EdgeApply) for v in edges):
                created_edges.extend(edges)

            if created_nodes and created_edges:
                raise ValueError(
                    "Cannot create both nodes and edges at the same time. Toolikt should call one at a time"
                )
            created_resources[Node.__name__].extend(created_nodes)
            created_resources[Edge.__name__].extend(created_edges)

            node_list = []
            if isinstance(nodes, Sequence):
                node_list.extend(nodes)
            elif isinstance(nodes, NodeApply):
                node_list.append(nodes)

            edge_list = []
            if isinstance(edges, Sequence):
                edge_list.extend(edges)
            elif isinstance(edges, EdgeApply):
                edge_list.append(edges)

            return InstancesApplyResult(
                nodes=NodeApplyResultList(
                    [
                        NodeApplyResult(
                            space=node.space,
                            external_id=node.external_id,
                            version=node.existing_version or 1,
                            was_modified=True,
                            last_updated_time=1,
                            created_time=1,
                        )
                        for node in node_list
                    ]
                ),
                edges=EdgeApplyResultList(
                    [
                        EdgeApplyResult(
                            space=edge.space,
                            external_id=edge.external_id,
                            version=edge.existing_version or 1,
                            was_modified=True,
                            last_updated_time=1,
                            created_time=1,
                        )
                        for edge in edge_list
                    ]
                ),
            )

        def create_nodes(
            nodes: NodeApply | Sequence[NodeApply] | None = None,
            **_,
        ) -> NodeApplyResult | NodeApplyResultList:
            """Create nodes in the mock client."""
            output = create_instances(nodes=nodes, edges=None)
            if isinstance(nodes, NodeApply):
                return output.nodes[0]
            return output.nodes

        def create_extraction_pipeline_config(config: ExtractionPipelineConfigWrite) -> ExtractionPipelineConfig:
            created_resources[resource_cls.__name__].append(config)
            return ExtractionPipelineConfig.load(config.dump(camel_case=True))

        def upload_bytes_files_api(content: str | bytes | TextIO | BinaryIO, **kwargs) -> FileMetadata:
            if not isinstance(content, bytes):
                raise NotImplementedError("Only bytes content is supported")
            if "id" not in kwargs:
                kwargs["id"] = len(calculate_hash(content, shorten=False))
            created_resources[resource_cls.__name__].append(
                {
                    **kwargs,
                }
            )
            return FileMetadata.load({to_camel_case(k): v for k, v in kwargs.items()})

        def upload_file_content_path_files_api(
            path: str,
            external_id: str | None = None,
            instance_id: NodeId | None = None,
        ) -> FileMetadata:
            return _upload_file_content_files_api(
                calculate_hash(Path(path), shorten=True), external_id=external_id, instance_id=instance_id
            )

        def upload_file_content_bytes_files_api(
            content: str | bytes,
            external_id: str | None = None,
            instance_id: NodeId | None = None,
        ) -> FileMetadata:
            if isinstance(content, bytes):
                try:
                    hashes: list[str] = []
                    with zipfile.ZipFile(io.BytesIO(content)) as z:
                        namelist = z.namelist()
                        for name in sorted(namelist):
                            hashes.append(calculate_hash(name, shorten=False))
                            if name.endswith("/"):
                                continue
                            with z.open(name) as f:
                                file_content = f.read()
                                hashes.append(calculate_hash(file_content, shorten=True))
                    result = calculate_hash(",".join(sorted(hashes)), shorten=True)
                except zipfile.BadZipFile:
                    result = calculate_hash(content, shorten=True)
            elif isinstance(content, str):
                result = calculate_hash(content, shorten=True)
            else:
                raise NotImplementedError("Only str and bytes content is supported")
            return _upload_file_content_files_api(result, external_id=external_id, instance_id=instance_id)

        def _upload_file_content_files_api(
            filehash: str,
            external_id: str | None = None,
            instance_id: NodeId | None = None,
        ) -> FileMetadata:
            if sum([bool(external_id), bool(instance_id)]) != 1:
                raise ValueError("Exactly one of external_id or instance_id must be set")

            if external_id:
                entry = {"external_id": external_id}
            else:
                entry = instance_id.dump(include_instance_type=False)
            entry["filehash"] = filehash

            created_resources[FileCRUD.__name__].append(entry)

            return FileMetadata(external_id, instance_id, id=len(filehash))

        def create_3dmodel(
            name: str, data_set_id: int | None = None, metadata: dict[str, str] | None = None
        ) -> ThreeDModel:
            created = ThreeDModel(name=name, data_set_id=data_set_id, metadata=metadata, created_time=1)
            created_resources[resource_cls.__name__].append(created)
            return created

        def apply_dml(
            id: dm.DataModelId,
            dml: str,
            name: str | None = None,
            description: str | None = None,
            previous_version: str | None = None,
            preserve_dml: bool | None = None,
        ) -> DMLApplyResult:
            created = GraphQLDataModelWrite(
                space=id.space,
                external_id=id.external_id,
                version=id.version,
                dml=dml,
                name=name,
                description=description,
                previous_version=previous_version,
                preserve_dml=preserve_dml,
            )
            created_resources[resource_cls.__name__].append(created)
            return DMLApplyResult(
                space=id.space,
                external_id=id.external_id,
                version=id.version,
                description=description,
                name=name,
                last_updated_time="1",
                created_time="1",
            )

        available_create_methods = {
            fn.__name__: fn
            for fn in [
                create_multiple,
                create_single,
                create_filemetadata,
                insert_dataframe,
                upload,
                upsert,
                create_instances,
                create_extraction_pipeline_config,
                upload_bytes_files_api,
                upload_file_content_path_files_api,
                upload_file_content_bytes_files_api,
                create_3dmodel,
                apply_dml,
                create_raw_table,
                create_nodes,
            ]
        }
        if mock_method not in available_create_methods:
            raise ValueError(
                f"Invalid mock create method {mock_method} for resource {resource_cls.__name__}. Supported {list(available_create_methods.keys())}"
            )
        method = available_create_methods[mock_method]
        return method

    def _create_retrieve_method(self, resource: APIResource, mock_method: str, client: CogniteClient) -> Callable:
        existing_resources = self._existing_resources
        resource_cls = resource.resource_cls
        read_list_cls = resource.list_cls

        def _create_verification_resource(*args, **kwargs):
            # Note that the written version of the resource does not contain the serves set variables,
            # so we need to set these manually
            if resource_cls is Space:
                ids = list(*args)
                spaces = [
                    Space(
                        space=space,
                        is_global=False,
                        last_updated_time=1,
                        created_time=1,
                    )
                    for space in ids
                ]
                return read_list_cls(spaces, cognite_client=client)
            elif resource_cls is DataSet:
                if "external_ids" in kwargs:
                    external_ids = kwargs["external_ids"]
                elif "external_id" in kwargs:
                    external_ids = [kwargs["external_id"]]
                else:
                    raise RuntimeError("No external_ids or external_id in kwargs")
                datasets = [
                    DataSet(
                        external_id=external_id,
                        name=external_id,
                        id=42,
                        last_updated_time=1,
                        created_time=1,
                    )
                    for external_id in external_ids
                ]
                return read_list_cls(datasets, cognite_client=client)
            elif resource_cls is ExtractionPipeline:
                if "external_ids" in kwargs:
                    external_ids = kwargs["external_ids"]
                elif "external_id" in kwargs:
                    external_ids = [kwargs["external_id"]]
                else:
                    raise RuntimeError("No external_ids or external_id in kwargs")
                pipelines = [
                    ExtractionPipeline(
                        external_id=external_id,
                        name=external_id,
                        data_set_id=42,
                        id=722,
                        last_updated_time=1,
                        created_time=1,
                    )
                    for external_id in external_ids
                ]
                return read_list_cls(pipelines, cognite_client=client)

            raise NotImplementedError(f"Return values not implemented for {resource_cls}")

        def return_values(*args, **kwargs):
            if self._return_verify_resources and resource_cls in {Space, DataSet, ExtractionPipeline}:
                return _create_verification_resource(*args, **kwargs)

            return read_list_cls(existing_resources[resource_cls.__name__], cognite_client=client)

        def return_data_models(
            ids: DataModelIdentifier | Sequence[DataModelIdentifier], inline_views: bool = False
        ) -> DataModelList:
            if not existing_resources[resource_cls.__name__]:
                return DataModelList([])
            id_set = {ids} if isinstance(ids, str | tuple | dm.DataModelId) else set(ids)
            to_return = read_list_cls([], cognite_client=client)
            for resource in existing_resources[resource_cls.__name__]:
                id_ = resource.as_id()
                if id_ in id_set or id_.as_tuple() in id_set or id_.as_tuple()[:2] in id_set:
                    to_return.append(resource)
            return to_return

        def iterate_values(*argd, **kwargs):
            list_ = return_values(*argd, **kwargs)
            return (value for value in list_)

        def return_instances(*args, **kwargs) -> InstancesResult:
            read_list = return_values(*args, **kwargs)
            return InstancesResult(nodes=read_list, edges=EdgeList([]))

        def return_value(*args, **kwargs):
            if value := existing_resources[resource_cls.__name__]:
                return read_list_cls(value, cognite_client=client)[0]
            elif self.return_verify_resources and resource_cls in {Space, DataSet, ExtractionPipeline}:
                return _create_verification_resource(*args, **kwargs)[0]
            else:
                return None

        def files_retrieve(id: int | None = None, external_id: str | None = None) -> FileMetadata:
            if id is not None:
                return FileMetadata(id=id, uploaded=True)
            elif external_id is not None:
                return FileMetadata(external_id=external_id, uploaded=True)
            else:
                return return_value(external_id=external_id)

        def data_model_retrieve(ids, *args, **kwargs):
            id_set = set(ids) if isinstance(ids, Sequence) else {ids}
            to_return = read_list_cls([], cognite_client=client)
            for resource in existing_resources[resource_cls.__name__]:
                id = resource.as_id()
                if id in id_set or (id.as_tuple() in id_set and id.as_tuple()[:2] in id_set):
                    to_return.append(resource)
            return to_return

        available_retrieve_methods = {
            fn.__name__: fn
            for fn in [
                return_values,
                return_value,
                data_model_retrieve,
                return_instances,
                files_retrieve,
                iterate_values,
                return_data_models,
            ]
        }
        if mock_method not in available_retrieve_methods:
            raise ValueError(
                f"Invalid mock retrieve method {mock_method} for resource {resource_cls.__name__}. Supported {available_retrieve_methods.keys()}"
            )
        method = available_retrieve_methods[mock_method]
        return method

    def _create_inspect_method(self, resource: APIResource, mock_method: str, client: CogniteClient) -> Callable:
        existing_resources = self._existing_resources
        resource_cls = resource.resource_cls

        def return_value(*args, **kwargs):
            if value := existing_resources[resource_cls.__name__]:
                return value[0]

            return TokenInspection(
                subject="test",
                projects=[ProjectSpec(url_name="test_project", groups=[123, 456])],
                capabilities=ProjectCapabilityList(
                    [
                        ProjectCapability(capability=capability, project_scope=AllProjectsScope())
                        for capability in _ALL_CAPABILITIES
                    ],
                    cognite_client=client,
                ),
            )

        available_inspect_methods = {
            fn.__name__: fn
            for fn in [
                return_value,
            ]
        }
        if mock_method not in available_inspect_methods:
            raise ValueError(
                f"Invalid mock retrieve method {mock_method} for resource {resource_cls.__name__}. Supported {available_inspect_methods.keys()}"
            )
        method = available_inspect_methods[mock_method]
        return method

    def _create_post_method(self, resource: APIResource, mock_method: str, client: CogniteClient) -> Callable:
        def post_method(
            url: str, json: dict[str, Any], params: dict[str, Any] | None = None, headers: dict[str, Any] | None = None
        ) -> Response:
            sessionResponse = Response()
            if url.endswith("/sessions"):
                sessionResponse.status_code = 200
                sessionResponse._content = b'{"items":[{"id":5192234284402249,"nonce":"QhlCnImCBwBNc72N","status":"READY","type":"ONESHOT_TOKEN_EXCHANGE"}]}'
            elif url.endswith("/functions/schedules"):
                sessionResponse.status_code = 201
                sessionResponse._content = str.encode(JSON.dumps(json))
            elif url.split("/")[-3] == "functions" and url.split("/")[-2].isdigit() and url.endswith("call"):
                sessionResponse.status_code = 201
                sessionResponse._content = str.encode(
                    JSON.dumps(FunctionCall(id=1, status="RUNNING").dump(camel_case=True))
                )
            else:
                raise ValueError(
                    f"The url {url} is called with post method, but not mocked. Please add in _create_post_method in approval.client.py"
                )
            return sessionResponse

        existing_resources = self._existing_resources
        resource_cls = resource.resource_cls

        def return_value(*args, **kwargs):
            return existing_resources[resource_cls.__name__][0]

        available_post_methods = {
            fn.__name__: fn
            for fn in [
                return_value,
                post_method,
            ]
        }
        if mock_method not in available_post_methods:
            raise ValueError(
                f"Invalid mock retrieve method {mock_method} for resource {resource_cls.__name__}. Supported {available_post_methods.keys()}"
            )
        method = available_post_methods[mock_method]
        return method

    def dump(self, sort: bool = True) -> dict[str, Any]:
        """This returns a dictionary with all the resources that have been created and deleted.

        The sorting is useful in snapshot testing, as it makes for a consistent output. If you want to check the order
        that the resources were created, you can set sort=False.

        Args:
            sort: If True, the resources will be sorted by externalId, dbName, name, or name[0] if externalId is not available.


        Returns:
            A dict with the resources that have been created and deleted, {resource_name: [resource, ...]}
        """
        dumped = {}
        if sort:
            created_resources = sorted(self.created_resources)
        else:
            created_resources = list(self.created_resources)
        for key in created_resources:
            values = self.created_resources[key]
            if values:
                dumped_resource = (value.dump(camel_case=True) if hasattr(value, "dump") else value for value in values)
                if sort:

                    def sort_key(v: dict[str, Any]) -> str:
                        for identifier_name in [
                            "externalId",
                            "external_id",
                            "dbName",
                            "space",
                            "name",
                            "workflowExternalId",
                        ]:
                            if identifier_name in v:
                                return v[identifier_name]
                        if "dbName" in v and "name" in v and isinstance(v["name"], list):
                            return v["dbName"] + "/" + v["name"][0]
                        if "transformationExternalId" in v and "destination" in v:
                            return v["transformationExternalId"] + v["destination"]
                        if "view" in v and "space" in v["view"] and "externalId" in v["view"]:
                            return v["view"]["space"] + "/" + v["view"]["externalId"]
                        raise ValueError(f"Could not find identifier in {v}")

                    dumped[key] = sorted(dumped_resource, key=sort_key)
                else:
                    dumped[key] = list(dumped_resource)

        # Standardize file paths
        for filemedata in dumped.get("FileMetadata", []):
            if "kwargs" in filemedata and "path" in filemedata["kwargs"] and "/" in filemedata["kwargs"]["path"]:
                filemedata["kwargs"]["path"] = filemedata["kwargs"]["path"].split("/")[-1]

        if self._deleted_resources:
            dumped["deleted"] = {}
            if sort:
                deleted_resources = sorted(self._deleted_resources)
            else:
                deleted_resources = list(self._deleted_resources)

            for key in deleted_resources:
                values = self._deleted_resources[key]

                def sort_deleted(x):
                    if not isinstance(x, dict):
                        return x
                    if "externalId" in x:
                        return x["externalId"]
                    if "db_name" in x and "name" in x and isinstance(x["name"], list):
                        return x["db_name"] + "/" + x["name"][0]
                    if "db_name" in x:
                        return x["db_name"]
                    return "missing"

                if values:
                    dumped["deleted"][key] = (
                        sorted(
                            values,
                            key=sort_deleted,
                        )
                        if sort
                        else list(values)
                    )

        return dumped

    def created_resources_of_type(self, resource_type: type[T_CogniteResource]) -> list[T_CogniteResource]:
        """This returns all the resources that have been created of a specific type.

        Args:
            resource_type: The type of resource to return, for example, 'TimeSeries', 'DataSet', 'Transformation'

        Returns:
            A list of all the resources that have been created of a specific type.
        """
        return self.created_resources.get(resource_type.__name__, [])

    def create_calls(self) -> dict[str, int]:
        """This returns all the calls that have been made to the mock client to create methods.

        For example, if you have mocked the 'time_series' API, and the code you test calls the 'time_series.create' method,
        then this method will return {'time_series': 1}
        """
        return {
            key: call_count
            for key, methods in self._create_methods.items()
            if (call_count := sum(method.call_count for method in methods))
        }

    def retrieve_calls(self) -> dict[str, int]:
        """This returns all the calls that have been made to the mock client to retrieve methods.

        For example, if you have mocked the 'time_series' API, and the code you test calls the 'time_series.list' method,
        then this method will return {'time_series': 1}
        """
        return {
            key: call_count
            for key, methods in self._retrieve_methods.items()
            if (call_count := sum(method.call_count for method in methods))
        }

    def inspect_calls(self) -> dict[str, int]:
        """This returns all the calls that have been made to the mock client to the inspect method."""
        return {
            key: call_count
            for key, methods in self._inspect_methods.items()
            if (call_count := sum(method.call_count for method in methods))
        }

    def delete_calls(self) -> dict[str, int]:
        """This returns all the calls that have been made to the mock client to delete methods.

        For example, if you have mocked the 'time_series' API, and the code you test calls the 'time_series.delete' method,
        then this method will return {'time_series': 1}
        """
        return {
            key: call_count
            for key, methods in self._delete_methods.items()
            if (call_count := sum(method.call_count for method in methods))
        }

    def not_mocked_calls(self) -> dict[str, int]:
        """This returns all the calls that have been made to the mock client to sub APIs that have not been mocked.

        For example, if you have not mocked the 'time_series' API, and the code you test calls the 'time_series.list' method,
        then this method will return {'time_series.list': 1}

        Returns:
            A dict with the calls that have been made to sub APIs that have not been mocked, {api_name.method_name: call_count}
        """
        mocked_apis: dict[str : set[str]] = defaultdict(set)
        for r in API_RESOURCES:
            if r.api_name.count(".") == 1:
                api_name, sub_api = r.api_name.split(".")
            elif r.api_name.count(".") == 0:
                api_name, sub_api = r.api_name, ""
            else:
                raise ValueError(f"Invalid api name {r.api_name}")
            mocked_apis[api_name] |= {sub_api} if sub_api else set()
        # These are mocked in the __init__ method
        for name, method in self.mock_client.lookup.__dict__.items():
            if not isinstance(method, MagicMock) or name.startswith("_") or name.startswith("assert_"):
                continue
            mocked_apis["lookup"].add(name)
        mocked_apis["verify"] = {"authorization"}
        mocked_apis["project"] = {"status"}

        not_mocked: dict[str, int] = defaultdict(int)
        for api_name, api in vars(self.mock_client).items():
            if not isinstance(api, MagicMock) or api_name.startswith("_") or api_name.startswith("assert_"):
                continue
            mocked_sub_apis = mocked_apis.get(api_name, set())
            for method_name in dir(api):
                if method_name.startswith("_") or method_name.startswith("assert_"):
                    continue
                method = getattr(api, method_name)
                if api_name not in mocked_apis and isinstance(method, MagicMock) and method.call_count:
                    not_mocked[f"{api_name}.{method_name}"] += method.call_count
                if hasattr(method, "_spec_set") and method._spec_set and method_name not in mocked_sub_apis:
                    # this is a sub api that must be checked
                    for sub_method_name in dir(method):
                        if sub_method_name.startswith("_") or sub_method_name.startswith("assert_"):
                            continue
                        sub_method = getattr(method, sub_method_name)
                        if isinstance(sub_method, MagicMock) and sub_method.call_count:
                            not_mocked[f"{api_name}.{method_name}.{sub_method_name}"] += sub_method.call_count
        # This is mocked in the __init__
        not_mocked.pop("iam.sessions.create", None)
        return dict(not_mocked)

    def auth_create_group_calls(self) -> Iterable[AuthGroupCalls]:
        groups = cast(GroupList, self.created_resources[Group.__name__])
        groups = sorted(groups, key=lambda x: x.name)
        for name, group in itertools.groupby(groups, key=lambda x: x.name):
            yield AuthGroupCalls(name=name, calls=list(group))



from datetime import datetime, timezone

import pytest
from cognite.client.data_classes import EventList, EventWrite, EventWriteList
from cognite.client.data_classes.data_modeling import EdgeId, NodeId

from cognite_toolkit._cdf_tk.client import ToolkitClient
from cognite_toolkit._cdf_tk.client.data_classes.canvas import (
    CanvasAnnotationApply,
    CanvasApply,
    ContainerReferenceApply,
    IndustrialCanvasApply,
)


@pytest.fixture(scope="session")
def three_events(toolkit_client: ToolkitClient) -> EventList:
    events = EventWriteList(
        [
            EventWrite(
                external_id=f"toolkit_test_canvas_event_{i}",
                type="CanvasEvent",
                description="This is a test event for the canvas.",
            )
            for i in range(3)
        ]
    )
    existing = toolkit_client.events.retrieve_multiple(external_ids=events.as_external_ids(), ignore_unknown_ids=True)
    if len(existing) == len(events):
        return existing
    existing_ids = existing.as_external_ids()
    created = toolkit_client.events.create([e for e in events if e.external_id not in existing_ids])
    existing.extend(created)
    return existing


def create_canvas(three_events: EventList) -> IndustrialCanvasApply:
    return IndustrialCanvasApply(
        canvas=CanvasApply(
            external_id="efc2de9d-27a5-4a3b-9779-dff11c572610",
            name="ToolkitTestData",
            created_by="ndTFZh9K9-m2W9WBKc30-Q",
            updated_at=datetime(2025, 6, 28, 10, 24, 34, tzinfo=timezone.utc),
            updated_by="ndTFZh9K9-m2W9WBKc30-Q",
            visibility="private",
            context=[{"type": "FILTERS", "payload": {"filters": []}}],
        ),
        annotations=[
            CanvasAnnotationApply(
                external_id="efc2de9d-27a5-4a3b-9779-dff11c572610_4fba01d2-bcb7-4871-934f-0ab7e754bce9",
                id_="4fba01d2-bcb7-4871-934f-0ab7e754bce9",
                annotation_type="polylineAnnotation",
                container_id="08ed536d-3d77-4692-80b2-d4a9ec6a3aea",
                is_selectable=True,
                is_draggable=True,
                is_resizable=True,
                properties_={
                    "style": {
                        "fill": "rgba(164,178,252,0.9)",
                        "stroke": "rgb(66, 85, 187)",
                        "opacity": 1,
                        "strokeWidth": 12,
                        "shouldEnableStrokeScale": True,
                    },
                    "zIndex": 3,
                    "vertices": [
                        {"x": -0.16666666666666666, "y": -0.27297185430463594},
                        {"x": 0.005, "y": 0.022971854304635456},
                    ],
                    "endEndType": "none",
                    "startEndType": "none",
                },
            ),
        ],
        container_references=[
            ContainerReferenceApply(
                external_id="efc2de9d-27a5-4a3b-9779-dff11c572610_5befa8e5-cdcf-4292-a6f9-ed176f9fb73c",
                container_reference_type="event",
                resource_id=three_events[0].id,
                id_="5befa8e5-cdcf-4292-a6f9-ed176f9fb73c",
                resource_sub_id=None,
                charts_id=None,
                label=three_events[0].external_id,
                properties_={"zIndex": 0, "unscaledWidth": 600, "unscaledHeight": 500},
                x=-287,
                y=-303,
                width=600,
                height=500,
                max_width=None,
                max_height=None,
            )
        ],
    )


class TestIndustrialCanvasAPI:
    def test_retrieve_non_existing(self, toolkit_client: ToolkitClient) -> None:
        result = toolkit_client.canvas.industrial.retrieve("non-existing-canvas")
        assert result is None, "Expected None when retrieving a non-existing canvas"

    def test_create_update_retrieve_delete(self, toolkit_client: ToolkitClient, three_events: EventList) -> None:
        canvas = create_canvas(three_events)

        deleted: list[NodeId | EdgeId] | None = None
        try:
            created = toolkit_client.canvas.industrial.create(canvas)
            assert set(created.as_ids()) == set(canvas.as_instance_ids(include_solution_tags=True))

            # Remove annotation
            canvas.annotations = []
            updated = toolkit_client.canvas.industrial.update(canvas)
            assert set(updated.as_ids()) == set(canvas.as_instance_ids(include_solution_tags=True))

            retrieved = toolkit_client.canvas.industrial.retrieve(canvas.as_id())

            assert retrieved.as_write().dump(keep_existing_version=False) == canvas.dump(keep_existing_version=False)

            deleted = toolkit_client.canvas.industrial.delete(canvas)

            assert toolkit_client.canvas.retrieve(canvas.as_id()) is None
        finally:
            if deleted is None:
                toolkit_client.data_modeling.instances.delete_fast(canvas.as_instance_ids())



# Copyright 2023 Cognite AS
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from pathlib import Path
from typing import final

from rich.console import Console

from cognite_toolkit._cdf_tk.client import ToolkitClient

from .auth import GroupCRUD, SecurityCategoryCRUD
from .classic import AssetCRUD
from .data_organization import DataSetsCRUD
from .datamodel import SpaceCRUD
from .extraction_pipeline import ExtractionPipelineCRUD
from .location import LocationFilterCRUD
from .raw import RawDatabaseCRUD, RawTableCRUD
from .timeseries import TimeSeriesCRUD


@final
class GroupResourceScopedCRUD(GroupCRUD):
    dependencies = frozenset(
        {
            SpaceCRUD,
            DataSetsCRUD,
            ExtractionPipelineCRUD,
            TimeSeriesCRUD,
            SecurityCategoryCRUD,
            LocationFilterCRUD,
            AssetCRUD,
            RawDatabaseCRUD,
            RawTableCRUD,
        }
    )

    def __init__(self, client: ToolkitClient, build_dir: Path | None, console: Console | None):
        super().__init__(client, build_dir, console, "resource_scoped_only")

    @property
    def display_name(self) -> str:
        return "resource-scoped groups"



from collections.abc import Iterable
from pathlib import Path

import pytest

from cognite_toolkit._cdf_tk.constants import MODULES
from cognite_toolkit._cdf_tk.resource_classes.graphql_model import GraphQLDataModelYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.data import COMPLETE_ORG_ALPHA_FLAGS
from tests.test_unit.utils import find_resources


def invalid_test_cases() -> Iterable:
    yield pytest.param(
        {
            "space": "my_space",
            "externalId": "my_model",
        },
        {"Missing required field: 'version'"},
        id="missing-required-field",
    )
    yield pytest.param(
        {
            "space": "",  # Empty space
            "externalId": "my_model",
            "version": "#1",
        },
        {
            "In field space string should have at least 1 character",
            "In field version string should match pattern '^[a-zA-Z0-9]([.a-zA-Z0-9_-]{0,41}[a-zA-Z0-9])?$'",
        },
        id="invalid space and version",
    )


class TestGraphQLDataModelYAML:
    @pytest.mark.parametrize("data", list(find_resources("GraphQLSchema", base=COMPLETE_ORG_ALPHA_FLAGS / MODULES)))
    def test_load_valid(self, data: dict[str, object]) -> None:
        loaded = GraphQLDataModelYAML.model_validate(data)

        assert loaded.model_dump(exclude_unset=True, by_alias=True) == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_test_cases()))
    def test_invalid_model_error_messages(self, data: dict | list, expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, GraphQLDataModelYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors



import itertools
from collections.abc import Iterator

import httpx
import pytest
import requests
import responses
import respx
from cognite.client.data_classes.capabilities import (
    DataModelInstancesAcl,
    DataModelsAcl,
    FilesAcl,
    TimeSeriesAcl,
)
from cognite.client.data_classes.data_modeling import (
    Container,
    DataModel,
    Edge,
    Node,
    NodeId,
    NodeList,
    Space,
    View,
)
from cognite.client.data_classes.data_modeling.cdm.v1 import CogniteFile, CogniteTimeSeries
from cognite.client.data_classes.data_modeling.statistics import SpaceStatistics

from cognite_toolkit._cdf_tk.client import ToolkitClient, ToolkitClientConfig
from cognite_toolkit._cdf_tk.client.data_classes.extended_filemetadata import (
    ExtendedFileMetadata,
)
from cognite_toolkit._cdf_tk.client.data_classes.extended_timeseries import ExtendedTimeSeries
from cognite_toolkit._cdf_tk.commands import PurgeCommand
from cognite_toolkit._cdf_tk.storageio.selectors import InstanceViewSelector, SelectedView
from tests.test_unit.utils import FakeCogniteResourceGenerator


@pytest.fixture(scope="session")
def cognite_timeseries_2000_list() -> NodeList[CogniteTimeSeries]:
    return NodeList[CogniteTimeSeries](
        [
            CogniteTimeSeries(
                space="test_space",
                external_id=f"test_timeseries_{i}",
                version=1,
                last_updated_time=1,
                created_time=1,
                is_step=True,
                time_series_type="numeric",
            )
            for i in range(2000)
        ]
    )


@pytest.fixture(scope="session")
def cognite_files_2000_list() -> NodeList[CogniteFile]:
    return NodeList[CogniteFile](
        [
            CogniteFile(
                space="test_space",
                external_id=f"test_file_{i}",
                version=1,
                last_updated_time=1,
                created_time=1,
                name=f"test_file_{i}.txt",
                mime_type="text/plain",
            )
            for i in range(2000)
        ]
    )


@pytest.fixture()
def timeseries_by_node_id(
    cognite_timeseries_2000_list: NodeList[CogniteTimeSeries],
) -> dict[NodeId, ExtendedTimeSeries]:
    return {
        ts.as_id(): ExtendedTimeSeries(
            id=i,
            external_id=ts.external_id,
            instance_id=ts.as_id(),
            is_string=ts.time_series_type == "string",
            is_step=ts.is_step,
            pending_instance_id=ts.as_id(),
        )
        for i, ts in enumerate(cognite_timeseries_2000_list)
    }


@pytest.fixture()
def files_by_node_id(
    cognite_files_2000_list: NodeList[CogniteFile],
) -> dict[NodeId, ExtendedFileMetadata]:
    return {
        file.as_id(): ExtendedFileMetadata(
            id=i,
            external_id=file.external_id,
            instance_id=file.as_id(),
            pending_instance_id=file.as_id(),
        )
        for i, file in enumerate(cognite_files_2000_list)
    }


@pytest.fixture()
def purge_responses(
    rsps: responses.RequestsMock, toolkit_config: ToolkitClientConfig
) -> Iterator[responses.RequestsMock]:
    config = toolkit_config
    rsps.add(
        responses.GET,
        f"{config.base_url}/api/v1/token/inspect",
        json={
            "subject": "123",
            "projects": [],
            "capabilities": [
                {
                    "projectScope": {"allProjects": {}},
                    **TimeSeriesAcl(
                        actions=[TimeSeriesAcl.Action.Read, TimeSeriesAcl.Action.Write],
                        scope=TimeSeriesAcl.Scope.All(),
                    ).dump(),
                },
                {
                    "projectScope": {"allProjects": {}},
                    **DataModelsAcl(
                        actions=[DataModelsAcl.Action.Read, DataModelsAcl.Action.Write], scope=DataModelsAcl.Scope.All()
                    ).dump(),
                },
                {
                    "projectScope": {"allProjects": {}},
                    **DataModelInstancesAcl(
                        actions=[DataModelInstancesAcl.Action.Read, DataModelInstancesAcl.Action.Write],
                        scope=DataModelInstancesAcl.Scope.All(),
                    ).dump(),
                },
                {
                    "projectScope": {"allProjects": {}},
                    **FilesAcl(
                        actions=[FilesAcl.Action.Read, FilesAcl.Action.Write], scope=FilesAcl.Scope.All()
                    ).dump(),
                },
            ],
        },
    )
    yield rsps


@pytest.fixture()
def purge_client(toolkit_config: ToolkitClientConfig) -> Iterator[ToolkitClient]:
    config = toolkit_config
    client = ToolkitClient(config, enable_set_pending_ids=True)
    yield client


class TestPurgeInstances:
    @pytest.mark.parametrize(
        "dry_run,unlink,instance_type",
        [
            pytest.param(True, True, "timeseries", id="dry run with unlink timeseries"),
            pytest.param(True, True, "files", id="dry run with unlink files"),
            pytest.param(True, False, "timeseries", id="dry run without unlink timeseries"),
            pytest.param(True, False, "files", id="dry run without unlink files"),
            pytest.param(False, True, "timeseries", id="purge with unlink timeseries"),
            pytest.param(False, True, "files", id="purge with unlink files"),
            pytest.param(False, False, "timeseries", id="purge without unlink timeseries"),
            pytest.param(False, False, "files", id="purge without unlink files"),
        ],
    )
    def test_purge(
        self,
        dry_run: bool,
        unlink: bool,
        instance_type: str,
        purge_client: ToolkitClient,
        purge_responses: responses.RequestsMock,
        respx_mock: respx.MockRouter,
        cognite_timeseries_2000_list: NodeList[CogniteTimeSeries],
        timeseries_by_node_id: dict[NodeId, ExtendedTimeSeries],
        cognite_files_2000_list: NodeList[CogniteFile],
        files_by_node_id: dict[NodeId, ExtendedFileMetadata],
    ) -> None:
        config = purge_client.config
        rsps = purge_responses
        instances = cognite_timeseries_2000_list if instance_type == "timeseries" else cognite_files_2000_list
        client = purge_client
        rsps.add(
            responses.POST,
            config.create_api_url("/models/instances/aggregate"),
            json={
                "items": [
                    {
                        "instanceType": "node",
                        "aggregates": [{"aggregate": "count", "value": len(instances), "property": "externalId"}],
                    }
                ]
            },
        )
        rsps.add(
            responses.POST,
            config.create_api_url("/models/instances/list"),
            json={"items": [instance.dump() for instance in instances]},
        )
        ts_objects = [ts.dump() for ts in timeseries_by_node_id.values()] if instance_type == "timeseries" else []
        file_objects = [file.dump() for file in files_by_node_id.values()] if instance_type == "files" else []
        if unlink:
            rsps.add(responses.POST, config.create_api_url("/timeseries/byids"), json={"items": ts_objects})
            rsps.add(responses.POST, config.create_api_url("/files/byids"), json={"items": file_objects})
        if unlink and not dry_run and instance_type == "timeseries":
            rsps.add(
                responses.POST,
                config.create_api_url("/timeseries/unlink-instance-ids"),
                json={"items": [ts.dump() for ts in timeseries_by_node_id.values()]},
            )
        if unlink and not dry_run and instance_type == "files":
            rsps.add(
                responses.POST,
                config.create_api_url("/files/unlink-instance-ids"),
                json={"items": [file.dump() for file in files_by_node_id.values()]},
            )
        if not dry_run:
            respx_mock.post(
                config.create_api_url("/models/instances/delete"),
            ).mock(
                return_value=httpx.Response(
                    status_code=200,
                    json={"items": [instance.as_id().dump() for instance in instances]},
                )
            )

        cmd = PurgeCommand(silent=True)
        result = cmd.instances(
            client,
            InstanceViewSelector(view=SelectedView(space="cdf_cdm", external_id="CogniteTimeSeries", version="v1")),
            dry_run=dry_run,
            auto_yes=True,
            unlink=unlink,
            verbose=False,
        )
        assert result.deleted == 2000


class TestPurgeSpace:
    @pytest.mark.usefixtures("disable_gzip")
    @pytest.mark.parametrize(
        "dry_run, include_space,delete_datapoints,delete_file_content,",
        itertools.product([True, False], [True, False], [True, False], [True, False]),
    )
    def test_purge(
        self,
        dry_run: bool,
        include_space: bool,
        delete_datapoints: bool,
        delete_file_content: bool,
        purge_client: ToolkitClient,
        purge_responses: responses.RequestsMock,
        respx_mock: respx.MockRouter,
    ) -> None:
        config = purge_client.config
        space = "test_space"
        rsps = purge_responses
        container_count = 10
        view_count = 15
        data_model_count = 3
        edge_count = 40
        node_count = 50
        ts_count = 14
        file_count = 7
        rsps.add(
            responses.POST,
            config.create_api_url("/models/statistics/spaces/byids"),
            json={
                "items": SpaceStatistics(
                    space, container_count, view_count, data_model_count, edge_count, 0, node_count, 0
                ).dump()
            },
        )

        def delete_callback(request: httpx.Request) -> httpx.Response:
            return httpx.Response(200, content=request.content)

        def delete_space_callback(request: requests.PreparedRequest) -> tuple[int, dict[str, str], str]:
            return 200, {}, request.body

        if not dry_run:
            gen = FakeCogniteResourceGenerator(seed=42)
            space_obj = gen.create_instance(Space)
            space_obj.space = space
            delete_urls = [
                "/models/containers/delete",
                "/models/views/delete",
                "/models/datamodels/delete",
                "/models/instances/delete",
            ]
            for url in delete_urls:
                respx_mock.post(config.create_api_url(url)).mock(side_effect=delete_callback)
            if include_space:
                rsps.add_callback(
                    method=responses.POST,
                    url=config.create_api_url("/models/spaces/delete"),
                    callback=delete_space_callback,
                )

            list_calls = [
                ("/models/containers", responses.GET, Container, container_count),
                ("/models/views", responses.GET, View, view_count),
                ("/models/datamodels", responses.GET, DataModel, data_model_count),
                ("/models/instances/list", responses.POST, Edge, edge_count),
                ("/models/instances/list", responses.POST, Node, node_count),
            ]
            if include_space:
                list_calls.append(("/models/spaces/byids", responses.POST, Space, 1))
            nodes: list[Node] = []
            for url, method, cls_, count in list_calls:
                items = [gen.create_instance(cls_) for _ in range(count)]
                if issubclass(cls_, Node):
                    nodes.extend(items)
                rsps.add(
                    method=method,
                    url=config.create_api_url(url),
                    json={"items": [item.dump() for item in items]},
                )

            retrieve_calls = []
            if not delete_datapoints:
                retrieve_calls.append(("/timeseries/byids", ExtendedTimeSeries, ts_count, nodes[:ts_count]))
            if not delete_file_content:
                retrieve_calls.append(
                    ("/files/byids", ExtendedFileMetadata, file_count, nodes[ts_count : ts_count + file_count])
                )

            for url, cls_, count, resource_nodes in retrieve_calls:
                resource_list = [gen.create_instance(cls_).dump() for _ in range(count)]
                for ts, node in zip(resource_list, resource_nodes):
                    ts["instanceId"] = node.as_id().dump(include_instance_type=False)
                rsps.add(
                    method=responses.POST,
                    url=config.create_api_url(url),
                    json={"items": resource_list},
                )

        cmd = PurgeCommand(silent=True)
        results = cmd.space(
            purge_client,
            "test_space",
            include_space=include_space,
            delete_datapoints=delete_datapoints,
            delete_file_content=delete_file_content,
            dry_run=dry_run,
            auto_yes=True,
            verbose=False,
        )
        expected_node_count = (
            node_count
            - (0 if delete_datapoints or dry_run else ts_count)
            - (0 if delete_file_content or dry_run else file_count)
        )
        expected = {
            "containers": container_count,
            "views": view_count,
            "data models": data_model_count,
            "edges": edge_count,
            "nodes": expected_node_count,
        }
        if include_space:
            expected["spaces"] = 1

        assert {name: value.deleted for name, value in results.data.items()} == expected



from collections.abc import Iterable
from pathlib import Path
from typing import Any

import pytest

from cognite_toolkit._cdf_tk.resource_classes.infield_location_config import InfieldLocationConfigYAML
from cognite_toolkit._cdf_tk.tk_warnings.fileread import ResourceFormatWarning
from cognite_toolkit._cdf_tk.validation import validate_resource_yaml_pydantic
from tests.data import COMPLETE_ORG_ALPHA_FLAGS
from tests.test_unit.utils import find_resources


def invalid_test_cases() -> Iterable:
    yield pytest.param(
        {"rootLocationExternalId": "myLocation", "space": "my_space"},
        {"Missing required field: 'externalId'"},
        id="Missing required field: externalId",
    )
    yield pytest.param(
        {
            "externalId": "my_config",
            "space": "my_space",
            "unknownField": "invalid_value",
            "anotherUnknownField": 123,
            "featureToggles": {
                "threeD": True,
                "invalidToggle": "bad_value",
            },
        },
        {
            "In featureToggles unused field: 'invalidToggle'",
            "Unused field: 'anotherUnknownField'",
            "Unused field: 'unknownField'",
        },
        id="Multiple extra fields at different levels",
    )
    yield pytest.param(
        {
            "externalId": "my_config",
            "space": "my_space",
            "featureToggles": {
                "threeD": "not_a_boolean",
                "trends": 123,
                "observations": {
                    "isEnabled": "not_a_boolean",
                    "isWriteBackEnabled": ["invalid_type"],
                },
            },
            "accessManagement": {
                "templateAdmins": "should_be_a_list",
                "checklistAdmins": 456,
            },
        },
        {
            "In accessManagement.checklistAdmins input should be a valid list. Got 456.",
            "In accessManagement.templateAdmins input should be a valid list. Got 'should_be_a_list'.",
            "In featureToggles.observations.isEnabled input should be a valid boolean. "
            "Got 'not_a_boolean' of type str.",
            "In featureToggles.observations.isWriteBackEnabled input should be a valid "
            "boolean. Got ['invalid_type'] of type list.",
            "In featureToggles.threeD input should be a valid boolean. Got 'not_a_boolean' of type str.",
            "In featureToggles.trends input should be a valid boolean. Got 123 of type int.",
        },
        id="Multiple type mismatches across nested structures",
    )


class TestInfieldCDMv1YAML:
    @pytest.mark.parametrize(
        "data",
        list(find_resources("InFieldLocationConfig", resource_dir="cdf_applications", base=COMPLETE_ORG_ALPHA_FLAGS)),
    )
    def test_load_valid(self, data: dict[str, Any]) -> None:
        loaded = InfieldLocationConfigYAML.model_validate(data)

        dumped = loaded.model_dump(exclude_unset=True, by_alias=True)
        assert dumped == data

    @pytest.mark.parametrize("data, expected_errors", list(invalid_test_cases()))
    def test_invalid_error_messages(self, data: dict[str, Any], expected_errors: set[str]) -> None:
        warning_list = validate_resource_yaml_pydantic(data, InfieldLocationConfigYAML, Path("some_file.yaml"))
        assert len(warning_list) == 1
        format_warning = warning_list[0]
        assert isinstance(format_warning, ResourceFormatWarning)

        assert set(format_warning.errors) == expected_errors
